#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
# Copyright (c) 2020 whitequark@whitequark.org
#
# The ECP5's DCU parameters/signals/instance have been partially documented by whitequark
# as part of the Yumewatari project: https://github.com/whitequark/Yumewatari.
#
# Code based in part on ``litex`` and ``liteiclink``.
# SPDX-License-Identifier: BSD-3-Clause
""" Soft PIPE backend for the Lattice ECP5 SerDes. """


from amaranth import *
from amaranth.lib.cdc import FFSynchronizer

from .lfps         import LFPSSquareWaveGenerator, LFPSSquareWaveDetector
from ..pipe        import PIPEInterface


class ECP5SerDesPLLConfiguration:
    def __init__(self, refclk, refclk_freq, linerate):
        self.refclk = refclk
        self.config = self.compute_config(refclk_freq, linerate)

    @staticmethod
    def compute_config(refclk_freq, linerate):
        for mult in [8, 10, 16, 20, 25]:
            current_linerate = refclk_freq*mult
            if current_linerate == linerate:
                return {
                    "mult":       mult,
                    "refck_freq": refclk_freq,
                    "linerate":   linerate,
                }
        msg = "No config found for {:3.2f} MHz refclk / {:3.2f} Gbps linerate."
        raise ValueError(msg.format(refclk_freq/1e6, linerate/1e9))



class ECP5SerDesConfigInterface(Elaboratable):
    """ Module that interfaces with the ECP5's SerDes Client Interface (SCI). """

    def __init__(self, serdes):
        self._serdes = serdes

        #
        # I/O port
        #

        # Control interface.
        self.dual_sel = Signal()
        self.chan_sel = Signal()
        self.re       = Signal()
        self.we       = Signal()
        self.done     = Signal()
        self.adr      = Signal(6)
        self.dat_w    = Signal(8)
        self.dat_r    = Signal(8)

        # SCI interface.
        self.sci_rd    = Signal()
        self.sci_wrn   = Signal()
        self.sci_addr  = Signal(6)
        self.sci_wdata = Signal(8)
        self.sci_rdata = Signal(8)



    def elaborate(self, platform):
        m = Module()

        m.d.comb += [
            self.sci_wrn.eq(1),

            self.sci_addr.eq(self.adr),
            self.sci_wdata.eq(self.dat_w)
        ]

        with m.FSM(domain="pipe"):

            with m.State("IDLE"):
                m.d.comb += self.done.eq(1)

                with m.If(self.we):
                    m.next = "WRITE"
                with m.Elif(self.re):
                    m.d.comb += self.sci_rd.eq(1),
                    m.next = "READ"

            with m.State("WRITE"):
                m.d.comb += self.sci_wrn.eq(0)
                m.next = "IDLE"


            with m.State("READ"):
                m.d.comb += self.sci_rd.eq(1)
                m.d.pipe += self.dat_r.eq(self.sci_rdata)
                m.next = "IDLE"

        return m


class ECP5SerDesRegisterTranslator(Elaboratable):
    """ Interface that converts control signals into SerDes register reads and writes. """

    def __init__(self, serdes, sci):
        self._serdes = serdes
        self._sci    = sci

        #
        # I/O port
        #
        self.loopback    = Signal()
        self.rx_polarity = Signal()
        self.tx_idle     = Signal()
        self.tx_polarity = Signal()
        self.rx_termination = Signal()


    def elaborate(self, platform):
        m = Module()
        sci = self._sci

        first = Signal()
        data  = Signal(8)


        with m.FSM(domain="pipe"):

            with m.State("IDLE"):
                m.d.pipe += first.eq(1)
                m.next = "READ-CH_01"

            with m.State("READ-CH_01"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.re.eq(1),
                    sci.adr.eq(0x01),
                ]

                with m.If(~first & sci.done):
                    m.d.comb += sci.re.eq(0)
                    m.d.pipe += [
                        data.eq(sci.dat_r),
                        first.eq(1)
                    ]
                    m.next = "WRITE-CH_01"


            with m.State("WRITE-CH_01"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.we.eq(1),
                    sci.adr.eq(0x01),
                    sci.dat_w.eq(data),
                    sci.dat_w[0].eq(self.rx_polarity),
                    sci.dat_w[1].eq(self.tx_polarity),
                ]
                with m.If(~first & sci.done):
                    m.d.comb += sci.we.eq(0)
                    m.d.pipe += first.eq(1)
                    m.next = "READ-CH_02"

            with m.State("READ-CH_02"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.re.eq(1),
                    sci.adr.eq(0x02),
                ]

                with m.If(~first & sci.done):
                    m.d.comb += sci.re.eq(0)
                    m.d.pipe += data.eq(sci.dat_r)
                    m.next = "WRITE-CH_02"

            with m.State("WRITE-CH_02"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.we.eq(1),
                    sci.adr.eq(0x02),
                    sci.dat_w.eq(data),
                    sci.dat_w[6].eq(self.tx_idle),  # pcie_ei_en
                ]

                with m.If(~first & sci.done):
                    m.d.pipe += first.eq(1)
                    m.d.comb += sci.we.eq(0)
                    m.next = "READ-CH_15"

            with m.State("READ-CH_15"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.re.eq(1),
                    sci.adr.eq(0x15),
                ]

                with m.If(~first & sci.done):
                    m.d.pipe += first.eq(1)
                    m.d.comb += sci.re.eq(0)
                    m.d.pipe += data.eq(sci.dat_r)
                    m.next = "WRITE-CH_15"

            with m.State("WRITE-CH_15"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.we.eq(1),
                    sci.adr.eq(0x15),
                    sci.dat_w.eq(data),
                ]
                with m.If(self.loopback):
                    m.d.comb += sci.dat_w[0:6].eq(0b110010) # lb_ctl

                with m.If(~first & sci.done):
                    m.d.comb += sci.we.eq(0)
                    m.next = "READ-CH_17"

            with m.State("READ-CH_17"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.re.eq(1),
                    sci.adr.eq(0x17),
                ]

                with m.If(~first & sci.done):
                    m.d.pipe += first.eq(1)
                    m.d.comb += sci.re.eq(0)
                    m.d.pipe += data.eq(sci.dat_r)
                    m.next = "WRITE-CH_17"

            with m.State("WRITE-CH_17"):
                m.d.pipe += first.eq(0)
                m.d.comb += [
                    sci.chan_sel.eq(1),
                    sci.we.eq(1),
                    sci.adr.eq(0x17),
                    sci.dat_w.eq(data),
                ]
                m.d.comb += sci.dat_w[0:5].eq(Mux(self.rx_termination, 22, 0))

                with m.If(~first & sci.done):
                    m.d.comb += sci.we.eq(0)
                    m.next = "IDLE"

        return m



class ECP5SerDesEqualizerInterface(Elaboratable):
    """ Interface that controls the ECP5 SerDes' equalization settings via SCI.

    Currently takes full ownership of the SerDes Client Interface.

    This unit allows runtime changing of the SerDes' equalizer settings.

    Attributes
    ----------
    enable_equalizer: Signal(), input
        Assert to enable the SerDes' equalizer.
    equalizer_pole: Signal(4), input
        Selects the pole used for the input equalization; ostensibly shifting the knee on the
        linear equalizer. The meaning of these values are not documented by Lattice.
    equalizer_level: Signal(2), input
        Selects the equalizer's gain. 0 = 6dB, 1 = 9dB, 2 = 12dB, 3 = undocumented.
        Note that the value `3` is marked as "not used" in the SerDes manual; but then used anyway
        by Lattice's reference designs.
    """

    SERDES_EQUALIZATION_REGISTER = 0x19

    def __init__(self, sci, serdes_channel):
        self._sci     = sci
        self._channel = serdes_channel

        #
        # I/O port
        #
        self.enable_equalizer = Signal()
        self.equalizer_pole   = Signal(4)
        self.equalizer_level  = Signal(2)


    def elaborate(self, platform):
        m = Module()
        sci = self._sci

        # Build the value to be written into the SCI equalizer register.
        m.d.comb += [
            sci.dat_w[0]    .eq(self.enable_equalizer),
            sci.dat_w[1:5]  .eq(self.equalizer_pole),
            sci.dat_w[5:7]  .eq(self.equalizer_level),

            # Set up a write to the equalizer control register.
            sci.chan_sel   .eq(self._channel),
            sci.we         .eq(1),
            sci.adr        .eq(self.SERDES_EQUALIZATION_REGISTER),
        ]

        return m




class ECP5SerDesEqualizer(Elaboratable):
    """ Interface that controls the ECP5 SerDes' equalization settings via SCI.

    Currently takes full ownership of the SerDes Client Interface.

    Ideally, an analog-informed receiver equalization would occur during USB3 link training. However,
    we're at best a simulacrum of a USB3 PHY built on an undocumented SerDes; so we'll do the best we
    can by measuring 8b10b encoding errors and trying various equalization settings until we've "minimized"
    bit error rate.

    Attributes
    ----------
    train_equalizer: Signal(), input
        When high, this unit attempts to train the Rx linear equalizer in order to minimize errors.
        This should be held only when a spectrally-rich data set is present, such as a training sequence.

    encoding_error_detected: Signal(), input
        Strobe; should be high each time the SerDes encounters an 8b10b encoding error.
    """

    # We'll try each equalizer setting for ~1024 cycles.
    # This value could easily be higher; but the higher this goes, the slower our counters
    # get; and we're operating in our fast, edge domain.
    CYCLES_PER_TRIAL = 127


    def __init__(self, sci, channel):
        self._sci     = sci
        self._channel = channel

        #
        # I/O port
        #
        self.train_equalizer         = Signal()
        self.encoding_error_detected = Signal()


    def elaborate(self, platform):
        m = Module()

        #
        # Equalizer interface.
        #
        m.submodules.interface = interface = ECP5SerDesEqualizerInterface(
            sci=self._sci,
            serdes_channel=self._channel
        )

        #
        # Bit error counter.
        #
        clear_errors    = Signal()
        bit_errors_seen = Signal(range(self.CYCLES_PER_TRIAL + 1))

        with m.If(clear_errors):
            m.d.pipe += bit_errors_seen.eq(0)
        with m.Elif(self.encoding_error_detected):
            m.d.pipe += bit_errors_seen.eq(bit_errors_seen + 1)


        #
        # Naive equalization trainer.
        #

        # We'll use the naive-est possible algorithm: we'll try every setting and see what
        # minimizes bit error rate. This could definitely be improved upon, but without documentation
        # for the equalizer, we're best going for an exhaustive approach.

        # We'll track six bits, as we have four bits of pole and two bits of gain we want to try.
        current_settings = Signal(6)
        m.d.comb += [
            interface.enable_equalizer                                .eq(1),
            Cat(interface.equalizer_level, interface.equalizer_pole)  .eq(current_settings)
        ]

        # Keep track of the best equalizer setting seen thus far.
        best_equalizer_setting = Signal.like(current_settings)
        best_bit_error_count   = Signal.like(bit_errors_seen)

        # Keep track of how long we've been in this trial.
        cycles_spent_in_trial  = Signal(range(self.CYCLES_PER_TRIAL))


        # If we're actively training the equalizer...
        with m.If(self.train_equalizer):
            m.d.pipe += cycles_spent_in_trial.eq(cycles_spent_in_trial + 1)

            # If we're finishing a trial...
            with m.If(cycles_spent_in_trial == (self.CYCLES_PER_TRIAL - 1)):

                # ... clear our error count...
                m.d.comb += clear_errors.eq(1)

                # ... move to the next set of settings ...
                m.d.pipe += current_settings.eq(current_settings + 1)

                # ... and if this is a new best, store it.
                with m.If(bit_errors_seen < best_bit_error_count):
                    m.d.pipe += [
                        best_bit_error_count    .eq(bit_errors_seen),
                        best_equalizer_setting  .eq(current_settings)
                    ]

        # If we're not currently in training, always apply our known best settings.
        with m.Else():
            m.d.pipe += current_settings.eq(best_equalizer_setting)


        return m


class ECP5SerDesResetSequencer(Elaboratable):
    """ Reset sequencer; ensures that the PLL, CDR, and PCS all start correctly. """

    RESET_CYCLES  = 8
    RX_LOS_CYCLES = 4000
    RX_LOL_CYCLES = 62500
    RX_ERR_CYCLES = 1024

    def __init__(self):

        #
        # I/O port
        #

        # Reset in.
        self.reset          = Signal()

        # Status in.
        self.tx_pll_locked  = Signal()
        self.rx_has_signal  = Signal()
        self.rx_cdr_locked  = Signal()
        self.rx_coding_err  = Signal()

        # Reset out.
        self.tx_pll_reset   = Signal(init=1)
        self.tx_pcs_reset   = Signal(init=1)
        self.rx_cdr_reset   = Signal(init=1)
        self.rx_pcs_reset   = Signal(init=1)

        # Status out.
        self.tx_pcs_ready   = Signal()
        self.rx_pcs_ready   = Signal()


    def elaborate(self, platform):
        m = Module()

        # Per [TN1261: "Reset Sequence"], the SerDes requires certain conditions to be met on start-up:
        # 1. The TxPLL must be locked before the CDR PLL reset is released.
        # 2. The PCS reset must be released only when the PLLs are locked.
        #
        # Per [Lattice FAQ 697], if the PCS is not held in reset and the Rx recovered clock is unstable,
        # the PCS FIFO pointers may get corrupted. In this case, there will be no 8b10b code violations
        # or disparity errors indicated, and the Rx link state machine will indicate synchronization
        # to the Rx bitstream, but the PCS may return corrupt data.
        #
        # Per [FPGA-PB-02001], the Rx CDR and PCS require a cascaded reset sequence for reliable operation.
        # On loss of signal, the Rx CDR and PCS are reset; and on loss of lock, the Rx PCS is reset.


        # Synchronize status signals to our clock.
        tx_pll_locked = Signal()
        rx_has_signal = Signal()
        rx_cdr_locked = Signal()
        rx_coding_err = Signal()
        m.submodules += [
            FFSynchronizer(self.tx_pll_locked, tx_pll_locked, o_domain="ss"),
            FFSynchronizer(self.rx_has_signal, rx_has_signal, o_domain="ss"),
            FFSynchronizer(self.rx_cdr_locked, rx_cdr_locked, o_domain="ss"),
            FFSynchronizer(self.rx_coding_err, rx_coding_err, o_domain="ss"),
        ]


        def apply_resets(m, tx_pll, tx_pcs, rx_cdr, rx_pcs):
            # The SerDes reset inputs are asynchronous; register our outputs so they do not have glitches.
            m.d.ss += [
                self.tx_pll_reset.eq(tx_pll),
                self.tx_pcs_reset.eq(tx_pcs),
                self.rx_cdr_reset.eq(rx_cdr),
                self.rx_pcs_reset.eq(rx_pcs),
            ]

        def apply_readys(m, tx_pcs, rx_pcs):
            m.d.ss += [
                self.tx_pcs_ready.eq(tx_pcs),
                self.rx_pcs_ready.eq(rx_pcs),
            ]


        timer = Signal(range(max(self.RESET_CYCLES, self.RX_LOS_CYCLES, self.RX_LOL_CYCLES)))

        with m.FSM(domain="ss"):

            # Hold everything in reset, initially.
            with m.State("INITIAL_RESET"):
                apply_resets(m, tx_pll=1, tx_pcs=1, rx_cdr=1, rx_pcs=1)
                apply_readys(m, tx_pcs=0, rx_pcs=0)

                m.next = "WAIT_FOR_TXPLL_LOCK"

            # Deassert Tx PLL reset, and wait for it to start up.
            with m.State("WAIT_FOR_TXPLL_LOCK"):
                apply_resets(m, tx_pll=0, tx_pcs=1, rx_cdr=1, rx_pcs=1)
                apply_readys(m, tx_pcs=0, rx_pcs=0)

                with m.If(tx_pll_locked):
                    m.d.ss += timer.eq(0)
                    m.next = "APPLY_TXPCS_RESET"

            # Reset Tx PCS.
            with m.State("APPLY_TXPCS_RESET"):
                apply_resets(m, tx_pll=0, tx_pcs=1, rx_cdr=1, rx_pcs=1)
                apply_readys(m, tx_pcs=0, rx_pcs=0)

                with m.If(timer + 1 != self.RESET_CYCLES):
                    m.d.ss += timer.eq(timer + 1)
                with m.Else():
                    m.d.ss += timer.eq(0)
                    m.next = "WAIT_FOR_RX_SIGNAL"

            # Deassert Tx PCS reset, and wait until Rx signal is present.
            with m.State("WAIT_FOR_RX_SIGNAL"):
                # CDR reset implies LOS reset; and must be deasserted for LOS to go low.
                # This is not documented in [TN1261], and contradicts [FPGA-PB-02001].
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=0, rx_pcs=1)
                apply_readys(m, tx_pcs=1, rx_pcs=0)

                with m.If(~rx_has_signal):
                    m.d.ss += timer.eq(0)
                with m.Else():
                    with m.If(timer + 1 != self.RX_LOS_CYCLES):
                        m.d.ss += timer.eq(timer + 1)
                    with m.Else():
                        m.d.ss += timer.eq(0)
                        m.next = "APPLY_CDR_RESET"

                with m.If(~tx_pll_locked):
                    m.next = "WAIT_FOR_TXPLL_LOCK"

            # Reset CDR.
            with m.State("APPLY_CDR_RESET"):
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=1, rx_pcs=1)
                apply_readys(m, tx_pcs=1, rx_pcs=0)

                with m.If(timer + 1 != self.RESET_CYCLES):
                    m.d.ss += timer.eq(timer + 1)
                with m.Else():
                    m.d.ss += timer.eq(0)
                    m.next = "DELAY_FOR_CDR_LOCK"

            # Deassert CDR reset, and wait until CDR had some time to lock (to embedded Rx clock).
            with m.State("DELAY_FOR_CDR_LOCK"):
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=0, rx_pcs=1)
                apply_readys(m, tx_pcs=1, rx_pcs=0)

                with m.If(timer + 1 != self.RX_LOL_CYCLES):
                    m.d.ss += timer.eq(timer + 1)
                with m.Else():
                    m.d.ss += timer.eq(0)
                    m.next = "CHECK_FOR_CDR_LOCK"

                with m.If(~rx_has_signal):
                    m.next = "WAIT_FOR_RX_SIGNAL"
                with m.If(~tx_pll_locked):
                    m.next = "WAIT_FOR_TXPLL_LOCK"

            # Wait until CDR has been locked for a while; and if it lost lock, reset it.
            with m.State("CHECK_FOR_CDR_LOCK"):
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=0, rx_pcs=1)
                apply_readys(m, tx_pcs=1, rx_pcs=0)

                with m.If(~rx_cdr_locked):
                    m.d.ss += timer.eq(0)
                    m.next = "APPLY_CDR_RESET"
                with m.Else():
                    with m.If(timer + 1 != self.RX_LOL_CYCLES):
                        m.d.ss += timer.eq(timer + 1)
                    with m.Else():
                        m.d.ss += timer.eq(0)
                        m.next = "APPLY_RXPCS_RESET"

                with m.If(~rx_has_signal):
                    m.next = "WAIT_FOR_RX_SIGNAL"
                with m.If(~tx_pll_locked):
                    m.next = "WAIT_FOR_TXPLL_LOCK"

            # Reset Rx PCS.
            with m.State("APPLY_RXPCS_RESET"):
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=0, rx_pcs=1)
                apply_readys(m, tx_pcs=1, rx_pcs=0)

                with m.If(timer + 1 != self.RESET_CYCLES):
                    m.d.ss += timer.eq(timer + 1)
                with m.Else():
                    m.d.ss += timer.eq(0)
                    m.next = "DELAY_FOR_RXPCS_LOCK"

            # Deassert Rx PCS reset, and wait until PCS had some time to lock (to a K28.5 comma).
            with m.State("DELAY_FOR_RXPCS_LOCK"):
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=0, rx_pcs=0)
                apply_readys(m, tx_pcs=1, rx_pcs=0)

                with m.If(timer + 1 != self.RX_ERR_CYCLES):
                    m.d.ss += timer.eq(timer + 1)
                with m.Else():
                    m.d.ss += timer.eq(0)
                    m.next = "CHECK_FOR_RXPCS_LOCK"

                with m.If(~rx_has_signal):
                    m.next = "WAIT_FOR_RX_SIGNAL"
                with m.If(~tx_pll_locked):
                    m.next = "WAIT_FOR_TXPLL_LOCK"

            # Wait until Rx PCS has been locked for a while; and if it lost lock, reset it.
            with m.State("CHECK_FOR_RXPCS_LOCK"):
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=0, rx_pcs=0)
                apply_readys(m, tx_pcs=1, rx_pcs=0)

                with m.If(rx_coding_err):
                    m.d.ss += timer.eq(0)
                    m.next = "APPLY_RXPCS_RESET"
                with m.Else():
                    with m.If(timer + 1 != self.RX_ERR_CYCLES):
                        m.d.ss += timer.eq(timer + 1)
                    with m.Else():
                        m.d.ss += timer.eq(0)
                        m.next = "IDLE"

                with m.If(~rx_has_signal):
                    m.next = "WAIT_FOR_RX_SIGNAL"
                with m.If(~tx_pll_locked):
                    m.next = "WAIT_FOR_TXPLL_LOCK"

            # Everything is okay; monitor for errors, and restart the reset sequence if necessary.
            with m.State("IDLE"):
                apply_resets(m, tx_pll=0, tx_pcs=0, rx_cdr=0, rx_pcs=0)
                apply_readys(m, tx_pcs=1, rx_pcs=1)

                with m.If(rx_coding_err):
                    m.next = "APPLY_RXPCS_RESET"
                with m.If(~rx_has_signal):
                    m.next = "WAIT_FOR_RX_SIGNAL"
                with m.If(~tx_pll_locked):
                    m.next = "WAIT_FOR_TXPLL_LOCK"

        return ResetInserter({"ss": self.reset})(m)


class ECP5SerDes(Elaboratable):
    """ Abstraction layer for working with the ECP5 SerDes. """

    def __init__(self, pll_config, tx_pads, rx_pads, dual=0, channel=0):
        assert dual    in [0, 1]
        assert channel in [0, 1]

        self._pll           = pll_config
        self._tx_pads       = tx_pads
        self._rx_pads       = rx_pads
        self._dual          = dual
        self._channel       = channel

        # Since we run at the 5 GT/s data rate, we always operate with 2x gearing;
        # the ECP5 fabric is not fast enough to process this much data otherwise.
        self._io_words      = 2

        #
        # I/O ports.
        #

        # Interface clock.
        self.pclk           = Signal()

        # Reset sequencing.
        self.reset          = Signal()
        self.tx_ready       = Signal()
        self.rx_ready       = Signal()

        # Core Rx and Tx lines.
        self.tx_data        = Signal(self._io_words * 8)
        self.tx_datak       = Signal(self._io_words)
        self.rx_data        = Signal(self._io_words * 8)
        self.rx_datak       = Signal(self._io_words)

        # TX controls
        self.tx_polarity    = Signal()
        self.tx_elec_idle   = Signal()
        self.tx_gpio_en     = Signal()
        self.tx_gpio        = Signal()

        # RX controls
        self.rx_polarity    = Signal()
        self.rx_gpio        = Signal()
        self.rx_termination = Signal()

        # RX status
        self.rx_status      = Signal(3)


    def elaborate(self, platform):
        m = Module()


        # Internal signals.
        tx_clk_full = Signal()
        tx_clk_half = Signal()

        tx_lol      = Signal()
        tx_bus      = Signal(24)

        rx_los      = Signal()
        rx_lol      = Signal()
        rx_err      = Signal()
        rx_ctc_urun = Signal()
        rx_ctc_orun = Signal()
        rx_align    = Signal()
        rx_bus      = Signal(24)


        #
        # Clocking / reset control.
        #

        # The SerDes needs to be brought up gradually; we'll do that here.
        m.submodules.reset_sequencer = reset = ECP5SerDesResetSequencer()
        m.d.comb += [
            reset.reset         .eq(self.reset),
            reset.tx_pll_locked .eq(~tx_lol),
            reset.rx_has_signal .eq(~rx_los),
            reset.rx_cdr_locked .eq(~rx_lol),
            reset.rx_coding_err .eq(rx_err),
            self.tx_ready       .eq(reset.tx_pcs_ready),
            self.rx_ready       .eq(reset.rx_pcs_ready),
        ]

        # Generate the PIPE interface clock from the half rate transmit byte clock, and use it to drive
        # both the Tx and the Rx FIFOs, to bring both halves of the data bus to the same clock domain.
        # The recovered Rx clock will not match the generated Tx clock; use the full rate transmit byte
        # clock to drive the CTC FIFO in the SerDes, which will compensate for the difference.
        m.d.comb += [
            self.pclk           .eq(tx_clk_half),
        ]


        #
        # SerDes parameter control.
        #

        # Some of the SerDes parameters cannot be directly controlled with fabric signals, but have to
        # be configured through the SerDes client interface.
        m.submodules.sci = sci = ECP5SerDesConfigInterface(self)
        m.submodules.sci_trans = sci_trans = ECP5SerDesRegisterTranslator(self, sci)
        m.d.comb += [
            sci_trans.tx_polarity   .eq(self.tx_polarity),
            sci_trans.rx_polarity   .eq(self.rx_polarity),
            sci_trans.rx_termination.eq(self.rx_termination),
        ]


        #
        # Core SerDes instantiation.
        #
        serdes_params = dict(
            # Note that Lattice Diamond needs these parameters in their provided bases (and string lengths!).
            # Changing their bases will work with the open toolchain, but will make Diamond mad.

            # DCU — power management
            p_D_MACROPDB            = "0b1",
            p_D_IB_PWDNB            = "0b1",
            p_D_TXPLL_PWDNB         = "0b1",
            i_D_FFC_MACROPDB        = 1,

            # DCU — reset
            i_D_FFC_MACRO_RST       = self.reset,
            i_D_FFC_DUAL_RST        = self.reset,

            # DCU — clocking
            i_D_REFCLKI             = self._pll.refclk,
            o_D_FFS_PLOL            = tx_lol,
            p_D_REFCK_MODE          = {
                25: "0b100",
                20: "0b000",
                16: "0b010",
                10: "0b001",
                 8: "0b011"}[self._pll.config["mult"]],
            p_D_TX_MAX_RATE         = "5.0",    # 5.0 Gbps
            p_D_TX_VCO_CK_DIV       = {
                32: "0b111",
                16: "0b110",
                 8: "0b101",
                 4: "0b100",
                 2: "0b010",
                 1: "0b000"}[1],                # DIV/1
            p_D_BITCLK_LOCAL_EN     = "0b1",    # Use clock from local PLL

            # DCU — clock multiplier unit
            # begin undocumented (Clarity Designer values for 5 Gbps PCIe used)
            p_D_CMUSETBIASI         = "0b00",
            p_D_CMUSETI4CPP         = "0d4",
            p_D_CMUSETI4CPZ         = "0d3",
            p_D_CMUSETI4VCO         = "0b00",
            p_D_CMUSETICP4P         = "0b01",
            p_D_CMUSETICP4Z         = "0b101",
            p_D_CMUSETINITVCT       = "0b00",
            p_D_CMUSETISCL4VCO      = "0b000",
            p_D_CMUSETP1GM          = "0b000",
            p_D_CMUSETP2AGM         = "0b000",
            p_D_CMUSETZGM           = "0b100",
            # end undocumented

            # DCU — unknown
            # begin undocumented (Clarity Designer values for 5 Gbps PCIe used)
            p_D_PD_ISET             = "0b11",
            p_D_RG_EN               = "0b0",
            p_D_RG_SET              = "0b00",
            p_D_SETICONST_AUX       = "0b01",
            p_D_SETICONST_CH        = "0b10",
            p_D_SETIRPOLY_AUX       = "0b10",
            p_D_SETIRPOLY_CH        = "0b10",
            p_D_SETPLLRC            = "0d1",
            # end undocumented

            # CHX common ---------------------------------------------------------------------------
            # CHX — protocol
            p_CHX_PROTOCOL          = "G8B10B",
            p_CHX_PCIE_MODE         = "0b1",

            p_CHX_ENC_BYPASS        = "0b0",    # Use the 8b10b encoder
            p_CHX_DEC_BYPASS        = "0b0",    # Use the 8b10b decoder

            # CHX receive --------------------------------------------------------------------------
            # CHX RX — power management
            p_CHX_RPWDNB            = "0b1",
            i_CHX_FFC_RXPWDNB       = 1,

            # CHX RX — reset
            i_CHX_FFC_RRST          = reset.rx_cdr_reset,
            i_CHX_FFC_LANE_RX_RST   = reset.rx_pcs_reset | rx_ctc_urun | rx_ctc_orun,

            # CHX RX — input
            i_CHX_HDINP             = self._rx_pads.p,
            i_CHX_HDINN             = self._rx_pads.n,

            p_CHX_LDR_RX2CORE_SEL   = "0b1",            # Enables low-speed out-of-band input.
            o_CHX_LDR_RX2CORE       = self.rx_gpio,

            p_CHX_RTERM_RX          = {
                "5k-ohms":        "0d0",
                "80-ohms":        "0d1",
                "75-ohms":        "0d4",
                "70-ohms":        "0d6",
                "60-ohms":        "0d11",
                "50-ohms":        "0d19",
                "46-ohms":        "0d25",
                "wizard-50-ohms": "0d22"}["5k-ohms"], # Set via SCI
            p_CHX_RXTERM_CM         = "0b10",   # Terminate RX to GND
            p_CHX_RXIN_CM           = "0b11",   # Common mode feedback

            # CHX RX — equalizer
            p_D_REQ_ISET            = "0b011",  # Undocumented, needs to be 010 or 011
            p_CHX_REQ_EN            = "0b1",    # Enable equalizer
            p_CHX_REQ_LVL_SET       = "0b01",   # Equalizer attenuation, 9 dB
            p_CHX_RX_RATE_SEL       = "0d09",   # Equalizer pole position, values documented as "TBD"

            # CHX RX — clocking
            p_CHX_FF_RX_H_CLK_EN    = "0b0",    # disable DIV/2 output clock
            p_CHX_FF_RX_F_CLK_DIS   = "0b1",    # disable DIV/1 output clock
            p_CHX_SEL_SD_RX_CLK     = "0b0",    # FIFO write driven by CTC buffer read clock
            i_CHX_FF_EBRD_CLK       = tx_clk_full,
            p_CHX_RX_GEAR_MODE      = "0b1",    # 1:2 gearbox
            i_CHX_FF_RXI_CLK        = tx_clk_half,

            # CHX RX — clock and data recovery
            p_CHX_CDR_MAX_RATE      = "5.0",    # 5.0 Gbps
            i_CHX_RX_REFCLK         = self._pll.refclk,
            p_CHX_RX_DCO_CK_DIV     = {
                32: "0b111",
                16: "0b110",
                 8: "0b101",
                 4: "0b100",
                 2: "0b010",
                 1: "0b000"}[1],                # DIV/1

            # begin undocumented (Clarity Designer values for 5 Gbps PCIe used)
            p_CHX_DCOATDCFG         = "0b00",
            p_CHX_DCOATDDLY         = "0b00",
            p_CHX_DCOBYPSATD        = "0b1",
            p_CHX_DCOCALDIV         = "0b010",
            p_CHX_DCOCTLGI          = "0b011",
            p_CHX_DCODISBDAVOID     = "0b1",
            p_CHX_DCOFLTDAC         = "0b00",
            p_CHX_DCOFTNRG          = "0b001",
            p_CHX_DCOIOSTUNE        = "0b010",
            p_CHX_DCOITUNE          = "0b00",
            p_CHX_DCOITUNE4LSB      = "0b010",
            p_CHX_DCOIUPDNX2        = "0b1",
            p_CHX_DCONUOFLSB        = "0b101",
            p_CHX_DCOSCALEI         = "0b01",
            p_CHX_DCOSTARTVAL       = "0b010",
            p_CHX_DCOSTEP           = "0b11",
            p_CHX_BAND_THRESHOLD    = "0d0",
            p_CHX_AUTO_FACQ_EN      = "0b1",
            p_CHX_AUTO_CALIB_EN     = "0b1",
            p_CHX_CALIB_CK_MODE     = "0b1",
            p_D_DCO_CALIB_TIME_SEL  = "0b00",
            p_CHX_REG_BAND_OFFSET   = "0d0",
            p_CHX_REG_BAND_SEL      = "0d0",
            p_CHX_REG_IDAC_SEL      = "0d0",
            p_CHX_REG_IDAC_EN       = "0b0",
            # end undocumented

            # CHX RX — loss of signal
            # Undocumented values were taken from Clarity Designer output for 5 Gbps PCIe
            o_CHX_FFS_RLOS          = rx_los,
            p_CHX_RLOS_SEL          = "0b1",
            p_CHX_RX_LOS_EN         = "0b1",
            p_CHX_RX_LOS_LVL        = "0b100",  # Values documented as "TBD"
            p_CHX_RX_LOS_CEQ        = "0b11",   # Values documented as "TBD"
            p_CHX_RX_LOS_HYST_EN    = "0b0",
            p_CHX_PDEN_SEL          = "0b1",    # phase detector disabled on LOS

            # CHX RX — loss of lock
            o_CHX_FFS_RLOL          = rx_lol,
            # USB requires the use of spread spectrum clocking, modulating the frequency from
            # +0 to -5000 ppm of the base 5 GHz clock.
            p_D_CDR_LOL_SET         = "0b10",   # ±4000 ppm lock, ±7000 ppm unlock

            # CHX RX — comma alignment
            # In the User Configured mode (generic 8b10b), the link state machine must be disabled
            # using CHx_LSM_DISABLE, or the CHx_FFC_ENABLE_CGALIGN and CHx_FFC_CR_EN_BITSLIP inputs
            # will not affect the WA and CDR.
            p_CHX_LSM_DISABLE       = "0b1",
            # The CHx_FFC_ENABLE_CGALIGN input is edge-sensitive; a posedge enables the word aligner,
            # which, once it discovers a comma, configures the barrel shifter and disables itself.
            # A constant level on this input does not affect WA; neither does the CHx_ENABLE_CG_ALIGN
            # parameter.
            i_CHX_FFC_ENABLE_CGALIGN= rx_err,

            p_CHX_UDF_COMMA_MASK    = "0x3ff",  # compare all bits
            p_CHX_UDF_COMMA_A       = "0x17c",   # 0b0101_111100, K28.5 RD- 10b code
            p_CHX_UDF_COMMA_B       = "0x283",   # 0b1010_000011, K28.5 RD+ 10b code

            # CHX RX — clock tolerance compensation
            # Due to spread spectrum modulation, the USB 3 word clock is, on average, 2.5% slower
            # than the base 5 GHz line rate. Since the USB soft logic always runs at a fraction of
            # the base line rate, SKP ordered sets only need to be removed, and RX FIFO underrun
            # can be handled using clock enables alone.
            p_CHX_CTC_BYPASS        = "0b0",    # enable CTC FIFO
            p_CHX_MIN_IPG_CNT       = "0b00",   # minimum interpacket gap of 1X (multiplied by match length)
            p_CHX_MATCH_2_ENABLE    = "0b1",    # enable  2 character skip matching (using characters 3..4)
            p_CHX_MATCH_4_ENABLE    = "0b0",    # disable 4 character skip matching (using characters 1..4)
            p_CHX_CC_MATCH_1        = "0x13c",   # K28.1 1+8b code
            p_CHX_CC_MATCH_2        = "0x13c",   # K28.1 1+8b code
            p_CHX_CC_MATCH_3        = "0x13c",   # K28.1 1+8b code
            p_CHX_CC_MATCH_4        = "0x13c",   # K28.1 1+8b code
            p_D_LOW_MARK            = "0d4",    # CTC FIFO low  water mark (mean=8)
            p_D_HIGH_MARK           = "0d12",   # CTC FIFO high water mark (mean=8)

            # The CTC FIFO underrun and overrun flags are 'sticky'; once the condition occurs, the flag
            # remains set until the RX PCS is reset. This affects the PIPE RxStatus output as well; to
            # be PIPE-compliant, reset the PCS immediately on overrun/underrun.
            o_CHX_FFS_CC_UNDERRUN   = rx_ctc_urun,
            o_CHX_FFS_CC_OVERRUN    = rx_ctc_orun,

            # CHX RX — data
            **{"o_CHX_FF_RX_D_%d" % n: rx_bus[n] for n in range(len(rx_bus))},

            # CHX transmit -------------------------------------------------------------------------
            # CHX TX — power management
            p_CHX_TPWDNB            = "0b1",
            i_CHX_FFC_TXPWDNB       = 1,

            # CHX TX — reset
            i_D_FFC_TRST            = reset.tx_pll_reset,
            i_CHX_FFC_LANE_TX_RST   = reset.tx_pcs_reset,

            # CHX TX — output
            o_CHX_HDOUTP            = self._tx_pads.p,
            o_CHX_HDOUTN            = self._tx_pads.n,

            p_CHX_LDR_CORE2TX_SEL   = "0b0",            # Uses CORE2TX_EN to enable out-of-band output.
            i_CHX_LDR_CORE2TX       = self.tx_gpio,
            i_CHX_FFC_LDR_CORE2TX_EN= self.tx_gpio_en,

            p_CHX_RTERM_TX          = {
                "5k-ohms":        "0d0",
                "80-ohms":        "0d1",
                "75-ohms":        "0d4",
                "70-ohms":        "0d6",
                "60-ohms":        "0d11",
                "50-ohms":        "0d19",
                "46-ohms":        "0d25",
                "wizard-50-ohms": "0d19"}["50-ohms"],
            p_CHX_TXAMPLITUDE       = "0d1000", # 1000 mV

            # CHX TX — equalization
            p_CHX_TDRV_SLICE0_CUR   = "0b011",  # 400 uA
            p_CHX_TDRV_SLICE0_SEL   = "0b01",   # main data
            p_CHX_TDRV_SLICE1_CUR   = "0b000",  # 100 uA
            p_CHX_TDRV_SLICE1_SEL   = "0b00",   # power down
            p_CHX_TDRV_SLICE2_CUR   = "0b11",   # 3200 uA
            p_CHX_TDRV_SLICE2_SEL   = "0b01",   # main data
            p_CHX_TDRV_SLICE3_CUR   = "0b10",   # 2400 uA
            p_CHX_TDRV_SLICE3_SEL   = "0b01",   # main data
            p_CHX_TDRV_SLICE4_CUR   = "0b00",   # 800 uA
            p_CHX_TDRV_SLICE4_SEL   = "0b00",   # power down
            p_CHX_TDRV_SLICE5_CUR   = "0b00",   # 800 uA
            p_CHX_TDRV_SLICE5_SEL   = "0b00",   # power down

            # CHX TX — clocking
            p_CHX_FF_TX_F_CLK_DIS   = "0b0",    # enable  DIV/1 output clock
            o_CHX_FF_TX_F_CLK       = tx_clk_full,
            p_CHX_FF_TX_H_CLK_EN    = "0b1",    # enable  DIV/2 output clock
            o_CHX_FF_TX_PCLK        = tx_clk_half, # ff_tx_pclk feeds a pclk net, ff_tx_h_clk does not
            p_CHX_TX_GEAR_MODE      = "0b1",    # 1:2 gearbox
            i_CHX_FF_TXI_CLK        = tx_clk_half,

            # CHX TX — data
            **{"i_CHX_FF_TX_D_%d" % n: tx_bus[n] for n in range(len(tx_bus))},

            i_CHX_FFC_EI_EN         = self.tx_elec_idle & ~self.tx_gpio_en,

            # SCI interface ------------------------------------------------------------------------
            **{"i_D_SCIWDATA%d" % n: sci.sci_wdata[n] for n in range(8)},
            **{"i_D_SCIADDR%d"  % n: sci.sci_addr [n] for n in range(6)},
            **{"o_D_SCIRDATA%d" % n: sci.sci_rdata[n] for n in range(8)},
            i_D_SCIENAUX  = sci.dual_sel,
            i_D_SCISELAUX = sci.dual_sel,
            i_CHX_SCIEN   = sci.chan_sel,
            i_CHX_SCISEL  = sci.chan_sel,
            i_D_SCIRD     = sci.sci_rd,
            i_D_SCIWSTN   = sci.sci_wrn,
        )

        # Translate the 'CHX' string to the correct channel name in each of our SerDes parameters,
        # and create our SerDes instance.
        serdes_params = {k.replace("CHX", f"CH{self._channel}"):v for (k,v) in serdes_params.items()}
        m.submodules.serdes = serdes = Instance("DCUA", **serdes_params)

        # Bind our SerDes to the correct location inside the FPGA.
        serdes.attrs["LOC"] = "DCU{}".format(self._dual)
        serdes.attrs["CHAN"] = "CH{}".format(self._channel)

        # SerDes decodes invalid 10b symbols to 0xEE with control bit set, which is not a part
        # of the 8b10b encoding space. We use it to drive the comma aligner and reset sequencer.
        # This signal is registered so that it can be sampled from asynchronous domains.
        m.d.pipe += [
            rx_err.eq(rx_bus[8]  & (rx_bus[ 0: 8] == 0xee) |
                      rx_bus[20] & (rx_bus[12:20] == 0xee)),
        ]

        #
        # TX and RX datapaths
        #
        rx_status = [Signal(3) for _ in range(self._io_words)]

        m.d.comb += [
            # Grab our received data directly from our SerDes; modifying things to match the
            # SerDes Rx bus layout, which squishes status signals between our two geared words.
            self.rx_data[0: 8]  .eq(rx_bus[ 0: 8]),
            self.rx_data[8:16]  .eq(rx_bus[12:20]),
            self.rx_datak[0]    .eq(rx_bus[8]),
            self.rx_datak[1]    .eq(rx_bus[20]),
            rx_status[0]        .eq(rx_bus[ 9:12]),
            rx_status[1]        .eq(rx_bus[21:24]),

            # Stick the data we'd like to transmit into the SerDes; again modifying things to match
            # the transmit bus layout.
            tx_bus[ 0: 8]       .eq(self.tx_data[0: 8]),
            tx_bus[12:20]       .eq(self.tx_data[8:16]),
            tx_bus[8]           .eq(self.tx_datak[0]),
            tx_bus[20]          .eq(self.tx_datak[1]),
        ]

        # The SerDes is providing us with two RxStatus words, one per byte; but we emit only one
        # for the entire word. Combine the status conditions together according to their priorities.
        for rx_status_code in 0b011, 0b010, 0b001, 0b111, 0b110, 0b101, 0b100:
            with m.If(Cat(rx_status[i] == rx_status_code for i in range(self._io_words)).any()):
                m.d.comb += self.rx_status.eq(rx_status_code)


        return m


class ECP5SerDesPIPE(PIPEInterface, Elaboratable):
    """ Wrapper around the core ECP5 SerDes that adapts it to the PIPE interface.

    The implementation-dependent behavior of the standard PIPE signals is described below:

    width :
        Interface width. Always 2 symbols.
    clk :
        Reference clock for the PHY receiver and transmitter. Could be routed through fabric,
        or connected to the output of an ``EXTREFB`` block. Frequency must be one of 250 MHz,
        200 MHz, or 312.5 MHz.
    pclk :
        Clock for the PHY interface. Frequency is always 250 MHz.
    phy_mode :
        PHY operating mode. Only SuperSpeed USB mode is supported.
    elas_buf_mode :
        Elastic buffer mode. Only nominal half-full mode is supported.
    rate :
        Link signaling rate. Only 5 GT/s is supported.
    power_down :
        Power management mode. Only P0 is supported.
    tx_deemph :
        Transmitter de-emphasis level. Only TBD is supported.
    tx_margin :
        Transmitter voltage levels. Only TBD is supported.
    tx_swing :
        Transmitter voltage swing level. Only full swing is supported.
    tx_detrx_lpbk :
    tx_elec_idle :
        Transmit control signals. Loopback and receiver detection are not implemented.
    tx_compliance :
    tx_ones_zeroes :
    rx_eq_training :
        These inputs are not implemented.
    power_present :
        This output is not implemented. External logic may drive it if necessary.
    """

    def __init__(self, *, tx_pads, rx_pads, channel=0, dual=0, refclk_frequency):
        super().__init__(width=2)

        self._tx_pads                 = tx_pads
        self._rx_pads                 = rx_pads
        self._channel                 = channel
        self._dual                    = dual
        self._refclk_frequency        = refclk_frequency


    def elaborate(self, platform):
        m = Module()

        #
        # SerDes instantiation.
        #
        pll_config = ECP5SerDesPLLConfiguration(
            refclk      = self.clk,
            refclk_freq = self._refclk_frequency,
            linerate    = 5e9
        )
        m.submodules.serdes = serdes = ECP5SerDes(
            pll_config  = pll_config,
            tx_pads     = self._tx_pads,
            rx_pads     = self._rx_pads,
            channel     = self._channel,
        )

        # Our soft PHY includes some logic that needs to run synchronously to the PIPE clock; create
        # a local clock domain to drive it.
        m.domains.pipe = ClockDomain(local=True, async_reset=True)
        m.d.comb += [
            ClockSignal("pipe")     .eq(serdes.pclk),
        ]


        #
        # LFPS generation & detection.
        #
        m.submodules.lfps_generator = lfps_generator = LFPSSquareWaveGenerator(25e6, 250e6)
        m.submodules.lfps_detector  = lfps_detector  = LFPSSquareWaveDetector(250e6)
        m.d.comb += [
            serdes.tx_gpio_en       .eq(lfps_generator.tx_gpio_en),
            serdes.tx_gpio          .eq(lfps_generator.tx_gpio),
            lfps_detector.rx_gpio   .eq(serdes.rx_gpio),
        ]


        #
        # PIPE interface signaling.
        #
        m.d.comb += [
            serdes.reset            .eq(self.reset),
            self.pclk               .eq(serdes.pclk),

            serdes.tx_elec_idle     .eq(self.tx_elec_idle),
            serdes.rx_polarity      .eq(self.rx_polarity),
            serdes.rx_termination   .eq(self.rx_termination),
            lfps_generator.generate .eq(self.tx_detrx_lpbk & self.tx_elec_idle),

            self.phy_status         .eq(~serdes.tx_ready),
            self.rx_valid           .eq(serdes.rx_ready),
            self.rx_status          .eq(serdes.rx_status),
            self.rx_elec_idle       .eq(~lfps_detector.present),

            serdes.tx_data          .eq(self.tx_data),
            serdes.tx_datak         .eq(self.tx_datak),
            self.rx_data            .eq(serdes.rx_data),
            self.rx_datak           .eq(serdes.rx_datak),
        ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Interfaces to LUNA's PSRAM chips."""

from amaranth import Const, Signal, Module, Cat, Elaboratable, Record, ClockSignal, ResetSignal, Instance
from amaranth.hdl.rec import DIR_FANIN, DIR_FANOUT
from amaranth.lib.cdc import FFSynchronizer


class HyperBusPHY(Record):
    """ Record representing a 16-bit HyperBus interface for use with a 2:1 PHY module. """

    def __init__(self):
        super().__init__([
            ('clk_en', 1, DIR_FANOUT),
            ('dq', [
                ('i', 16, DIR_FANIN),
                ('o', 16, DIR_FANOUT),
                ('e', 1,  DIR_FANOUT),
            ]),
            ('rwds', [
                ('i', 2,  DIR_FANIN),
                ('o', 2,  DIR_FANOUT),
                ('e', 1,  DIR_FANOUT),
            ]),
            ('cs',     1, DIR_FANOUT),
            ('reset',  1, DIR_FANOUT)
        ])


class HyperRAMInterface(Elaboratable):
    """ Gateware interface to HyperRAM series self-refreshing DRAM chips.

    I/O port:
        B: phy              -- The primary physical connection to the DRAM chip.

        I: address[32]      -- The address to be targeted by the given operation.
        I: register_space   -- When set to 1, read and write requests target registers instead of normal RAM.
        I: perform_write    -- When set to 1, a transfer request is viewed as a write, rather than a read.
        I: single_page      -- If set, data accesses will wrap around to the start of the current page when done.
        I: start_transfer   -- Strobe that goes high for 1-8 cycles to request a read operation.
                               [This added duration allows other clock domains to easily perform requests.]
        I: final_word       -- Flag that indicates the current word is the last word of the transaction.

        O: read_data[16]    -- word that holds the 16 bits most recently read from the PSRAM
        I: write_data[16]   -- word that accepts the data to output during this transaction

        O: idle             -- High whenever the transmitter is idle (and thus we can start a new piece of data.)
        O: read_ready       -- Strobe that indicates when new data is ready for reading
        O: write_ready      -- Strobe that indicates `write_data` has been latched and is ready for new data
    """

    LOW_LATENCY_CLOCKS  = 7
    HIGH_LATENCY_CLOCKS = 14

    def __init__(self, *, phy):
        """
        Parmeters:
            phy           -- The RAM record that should be connected to this RAM chip.
        """

        #
        # I/O port.
        #
        self.phy              = phy

        # Control signals.
        self.address          = Signal(32)
        self.register_space   = Signal()
        self.perform_write    = Signal()
        self.single_page      = Signal()
        self.start_transfer   = Signal()
        self.final_word       = Signal()

        # Status signals.
        self.idle             = Signal()
        self.read_ready       = Signal()
        self.write_ready      = Signal()

        # Data signals.
        self.read_data        = Signal(16)
        self.write_data       = Signal(16)


    def elaborate(self, platform):
        m = Module()

        #
        # Latched control/addressing signals.
        #
        is_read         = Signal()
        is_register     = Signal()
        current_address = Signal(32)
        is_multipage    = Signal()

        #
        # FSM datapath signals.
        #

        # Tracks whether we need to add an extra latency period between our
        # command and the data body.
        extra_latency   = Signal()

        # Tracks how many cycles of latency we have remaining between a command
        # and the relevant data stages.
        latency_clocks_remaining  = Signal(range(0, self.HIGH_LATENCY_CLOCKS + 1))

        # Store last edge values of RWDS and DQ lines.
        # This is used to handle clock inversion cases.
        last_half_rwds = Signal()
        last_half_dq   = Signal(len(self.phy.dq.i)//2)

        #
        # Core operation FSM.
        #

        # Provide defaults for our control/status signals.
        m.d.sync += [
            self.phy.clk_en     .eq(1),
            self.phy.cs         .eq(1),
            self.phy.rwds.e     .eq(0),
            self.phy.dq.e       .eq(0),
        ]
        m.d.comb += self.write_ready.eq(0),

        # Commands, in order of bytes sent:
        #   - WRBAAAAA
        #     W         => selects read or write; 1 = read, 0 = write
        #      R        => selects register or memory; 1 = register, 0 = memory
        #       B       => selects burst behavior; 0 = wrapped, 1 = linear
        #        AAAAA  => address bits [27:32]
        #
        #   - AAAAAAAA  => address bits [19:27]
        #   - AAAAAAAA  => address bits [11:19]
        #   - AAAAAAAA  => address bits [ 3:16]
        #   - 00000000  => [reserved]
        #   - 00000AAA  => address bits [ 0: 3]
        ca = Signal(48)
        m.d.comb += ca.eq(Cat(
            current_address[0:3],
            Const(0, 13),
            current_address[3:32],
            is_multipage,
            is_register,
            is_read
        ))

        with m.FSM() as fsm:

            # IDLE state: waits for a transaction request
            with m.State('IDLE'):
                m.d.comb += self.idle        .eq(1)
                m.d.sync += self.phy.clk_en  .eq(0)

                # Once we have a transaction request, latch in our control
                # signals, and assert our chip-select.
                with m.If(self.start_transfer):
                    m.next = 'LATCH_RWDS'

                    m.d.sync += [
                        is_read             .eq(~self.perform_write),
                        is_register         .eq(self.register_space),
                        is_multipage        .eq(~self.single_page),
                        current_address     .eq(self.address),
                        self.phy.dq.o       .eq(0),
                    ]

                with m.Else():
                    m.d.sync += self.phy.cs.eq(0)


            # LATCH_RWDS -- latch in the value of the RWDS signal,
            # which determines our read/write latency.
            with m.State("LATCH_RWDS"):
                m.d.sync += extra_latency.eq(self.phy.rwds.i),
                m.d.sync += self.phy.clk_en.eq(0)
                m.next="SHIFT_COMMAND0"


            # SHIFT_COMMANDx -- shift each of our command words out
            with m.State('SHIFT_COMMAND0'):
                # Output our first byte of our command.
                m.d.sync += [
                    self.phy.dq.o.eq(ca[32:48]),
                    self.phy.dq.e.eq(1)
                ]
                m.next = 'SHIFT_COMMAND1'

            with m.State('SHIFT_COMMAND1'):
                m.d.sync += [
                    self.phy.dq.o.eq(ca[16:32]),
                    self.phy.dq.e.eq(1)
                ]
                m.next = 'SHIFT_COMMAND2'

            with m.State('SHIFT_COMMAND2'):
                m.d.sync += [
                    self.phy.dq.o.eq(ca[0:16]),
                    self.phy.dq.e.eq(1)
                ]

                # If we have a register write, we don't need to handle
                # any latency. Move directly to our SHIFT_DATA state.
                with m.If(is_register & ~is_read):
                    m.next = 'WRITE_DATA'

                # Otherwise, react with either a short period of latency
                # or a longer one, depending on what the RAM requested via
                # RWDS.
                with m.Else():
                    m.next = "HANDLE_LATENCY"

                    # FIXME: our HyperRAM part has a fixed latency, but we could need to detect 
                    # different variants from the configuration register in the future.
                    with m.If(extra_latency | 1):
                        m.d.sync += latency_clocks_remaining.eq(self.HIGH_LATENCY_CLOCKS-2)
                    with m.Else():
                        m.d.sync += latency_clocks_remaining.eq(self.LOW_LATENCY_CLOCKS-2)


            # HANDLE_LATENCY -- applies clock cycles until our latency period is over.
            with m.State('HANDLE_LATENCY'):
                m.d.sync += latency_clocks_remaining.eq(latency_clocks_remaining - 1)

                with m.If(latency_clocks_remaining == 0):
                    with m.If(is_read):
                        m.next = 'READ_DATA'
                    with m.Else():
                        m.next = 'WRITE_DATA'


            # READ_DATA -- reads words from the PSRAM
            with m.State('READ_DATA'):

                # Store data sampled in last edge.
                m.d.sync += [
                    last_half_rwds .eq(self.phy.rwds.i[0]),
                    last_half_dq   .eq(self.phy.dq.i[:8])
                ]

                # If RWDS has changed, the host has just sent us new data.
                # Sample it, and indicate that we now have a valid piece of new data.
                with m.If(self.phy.rwds.i == 0b10):
                    m.d.comb += [
                        self.read_data     .eq(self.phy.dq.i),
                        self.read_ready    .eq(1),
                    ]

                    # If our controller is done with the transaction, end it.
                    with m.If(self.final_word):
                        m.next = 'RECOVERY'

                # Manage clock inversion: the data is divided between the current cycle 
                # and the preceding one.
                with m.Elif(Cat(self.phy.rwds.i[1], last_half_rwds) == 0b10):
                    m.d.comb += [
                        self.read_data     .eq(Cat(self.phy.dq.i[8:], last_half_dq)),
                        self.read_ready    .eq(1),
                    ]

                    # If our controller is done with the transaction, end it.
                    with m.If(self.final_word):
                        m.next = 'RECOVERY'

            # WRITE_DATA -- write a word to the PSRAM
            with m.State("WRITE_DATA"):
                m.d.sync += [
                    self.phy.dq.o    .eq(self.write_data),
                    self.phy.dq.e    .eq(1),
                    self.phy.rwds.e  .eq(~is_register),
                    self.phy.rwds.o  .eq(0),
                ]
                m.d.comb += self.write_ready.eq(1),

                # If we just finished a register write, we're done -- there's no need for recovery.
                with m.If(is_register):
                    m.next = 'IDLE'

                with m.Elif(self.final_word):
                    m.next = 'RECOVERY'


            # RECOVERY state: wait for the required period of time before a new transaction
            with m.State('RECOVERY'):
                m.d.sync += [
                    self.phy.cs     .eq(0),
                    self.phy.clk_en .eq(0),
                    last_half_rwds  .eq(0),
                    last_half_dq    .eq(0),

                ]

                # TODO: implement recovery
                m.next = 'IDLE'



        return m


class HyperRAMPHY(Elaboratable):
    """ Gateware PHY for HyperRAM series self-refreshing DRAM chips.

    I/O port:
        B: bus              -- The primary physical connection to the DRAM chip.
    """

    def __init__(self, *, bus, in_skew=None, out_skew=None, clock_skew=None):
        self.bus = bus
        self.phy = HyperBusPHY()

    def elaborate(self, platform):
        m = Module()

        m.submodules += [
            FFSynchronizer(self.phy.cs,     self.bus.cs.o,    stages=3),
            FFSynchronizer(self.phy.rwds.e, self.bus.rwds.oe, stages=3),
            FFSynchronizer(self.phy.dq.e,   self.bus.dq.oe,   stages=3),
        ]

        # Clock
        clk_out = Signal()
        m.submodules += [
            Instance("ODDRX1F",
                i_D0=self.phy.clk_en,
                i_D1=0,
                i_SCLK=ClockSignal(),
                i_RST=ResetSignal(),
                o_Q=clk_out,
            ),
            Instance("DELAYF",
                i_A=clk_out,
                o_Z=self.bus.clk.o,
                # TODO: connect up move/load/dir
                p_DEL_VALUE=int(2e-9 / 25e-12),
            ),
        ]

        # RWDS out
        m.submodules += [
            Instance("ODDRX1F",
                i_D0=self.phy.rwds.o[1],
                i_D1=self.phy.rwds.o[0],
                i_SCLK=ClockSignal(),
                i_RST=ResetSignal(),
                o_Q=self.bus.rwds.o,
            ),
        ]

        # RWDS in
        rwds_in = Signal()
        m.submodules += [
            Instance("DELAYF",
                i_A=self.bus.rwds.i,
                o_Z=rwds_in,
                # TODO: connect up move/load/dir
                p_DEL_VALUE=int(0e-9 / 25e-12),
            ),
            Instance("IDDRX1F",
                i_D=rwds_in,
                i_SCLK=ClockSignal(),
                i_RST=ResetSignal(),
                o_Q0=self.phy.rwds.i[1],
                o_Q1=self.phy.rwds.i[0],
            ),
        ]

        # DQ
        for i in range(8):
            # Out
            m.submodules += [
                Instance("ODDRX1F",
                    i_D0=self.phy.dq.o[i+8],
                    i_D1=self.phy.dq.o[i],
                    i_SCLK=ClockSignal(),
                    i_RST=ResetSignal(),
                    o_Q=self.bus.dq.o[i],
                ),
            ]

            # In
            dq_in = Signal(name=f"dq_in{i}")
            m.submodules += [
                Instance("DELAYF",
                    i_A=self.bus.dq.i[i],
                    o_Z=dq_in,
                    # TODO: connect up move/load/dir
                    p_DEL_VALUE=int(0e-9 / 25e-12),
                ),
                Instance("IDDRX1F",
                    i_D=dq_in,
                    i_SCLK=ClockSignal(),
                    i_RST=ResetSignal(),
                    o_Q0=self.phy.dq.i[i+8],
                    o_Q1=self.phy.dq.i[i],
                ),
            ]

        return m


class HyperBusDQSPHY(Record):
    """ Record representing a 32-bit HyperBus interface on a DQS group for use with a 4:1 PHY module. """

    def __init__(self):
        super().__init__([
            ('clk_en', 2, DIR_FANOUT),
            ('dq', [
                ('i', 32, DIR_FANIN),
                ('o', 32, DIR_FANOUT),
                ('e', 1,  DIR_FANOUT),
            ]),
            ('rwds', [
                ('i', 4,  DIR_FANIN),
                ('o', 4,  DIR_FANOUT),
                ('e', 1,  DIR_FANOUT),
            ]),
            ('cs',        1, DIR_FANOUT),
            ('reset',     1, DIR_FANOUT),
            ('read',      2, DIR_FANIN),
            ('datavalid', 1, DIR_FANOUT),
            ('burstdet',  1, DIR_FANOUT)
        ])



class HyperRAMDQSInterface(Elaboratable):
    """ Gateware interface to HyperRAM series self-refreshing DRAM chips, using ECP5 DQS logic.

    I/O port:
        B: phy              -- The primary physical connection to the DRAM chip.

        I: address[32]      -- The address to be targeted by the given operation.
        I: register_space   -- When set to 1, read and write requests target registers instead of normal RAM.
        I: perform_write    -- When set to 1, a transfer request is viewed as a write, rather than a read.
        I: single_page      -- If set, data accesses will wrap around to the start of the current page when done.
        I: start_transfer   -- Strobe that goes high for 1-8 cycles to request a read operation.
                               [This added duration allows other clock domains to easily perform requests.]
        I: final_word       -- Flag that indicates the current word is the last word of the transaction.

        O: read_data[32]    -- word that holds the 32 bits most recently read from the PSRAM
        I: write_data[32]   -- word that accepts the data to output during this transaction

        O: idle             -- High whenever the transmitter is idle (and thus we can start a new piece of data.)
        O: read_ready       -- Strobe that indicates when new data is ready for reading
        O: write_ready      -- Strobe that indicates `write_data` has been latched and is ready for new data
    """

    LOW_LATENCY_CLOCKS  = 3
    HIGH_LATENCY_CLOCKS = 5

    def __init__(self, *, phy):
        """
        Parmeters:
            phy           -- The RAM record that should be connected to this RAM chip.
        """

        #
        # I/O port.
        #
        self.phy              = phy

        # Control signals.
        self.address          = Signal(32)
        self.register_space   = Signal()
        self.perform_write    = Signal()
        self.single_page      = Signal()
        self.start_transfer   = Signal()
        self.final_word       = Signal()

        # Status signals.
        self.idle             = Signal()
        self.read_ready       = Signal()
        self.write_ready      = Signal()

        # Data signals.
        self.read_data        = Signal(32)
        self.write_data       = Signal(32)


    def elaborate(self, platform):
        m = Module()

        #
        # Latched control/addressing signals.
        #
        is_read         = Signal()
        is_register     = Signal()
        current_address = Signal(32)
        is_multipage    = Signal()

        #
        # FSM datapath signals.
        #

        # Tracks whether we need to add an extra latency period between our
        # command and the data body.
        extra_latency   = Signal()

        # Tracks how many cycles of latency we have remaining between a command
        # and the relevant data stages.
        latency_clocks_remaining  = Signal(range(0, self.HIGH_LATENCY_CLOCKS + 1))

        #
        # Core operation FSM.
        #

        # Provide defaults for our control/status signals.
        m.d.sync += [
            self.phy.clk_en     .eq(0b11),
            self.phy.cs         .eq(1),
            self.phy.rwds.e     .eq(0),
            self.phy.dq.e       .eq(0),
            self.phy.read       .eq(0),
        ]
        m.d.comb += self.write_ready.eq(0),

        # Commands, in order of bytes sent:
        #   - WRBAAAAA
        #     W         => selects read or write; 1 = read, 0 = write
        #      R        => selects register or memory; 1 = register, 0 = memory
        #       B       => selects burst behavior; 0 = wrapped, 1 = linear
        #        AAAAA  => address bits [27:32]
        #
        #   - AAAAAAAA  => address bits [19:27]
        #   - AAAAAAAA  => address bits [11:19]
        #   - AAAAAAAA  => address bits [ 3:16]
        #   - 00000000  => [reserved]
        #   - 00000AAA  => address bits [ 0: 3]
        ca = Signal(48)
        m.d.comb += ca.eq(Cat(
            current_address[0:3],
            Const(0, 13),
            current_address[3:32],
            is_multipage,
            is_register,
            is_read
        ))

        with m.FSM() as fsm:

            # IDLE state: waits for a transaction request
            with m.State('IDLE'):
                m.d.comb += self.idle        .eq(1)
                m.d.sync += self.phy.clk_en  .eq(0)

                # Once we have a transaction request, latch in our control
                # signals, and assert our chip-select.
                with m.If(self.start_transfer):
                    m.next = 'LATCH_RWDS'

                    m.d.sync += [
                        is_read             .eq(~self.perform_write),
                        is_register         .eq(self.register_space),
                        is_multipage        .eq(~self.single_page),
                        current_address     .eq(self.address),
                        self.phy.dq.o       .eq(0),
                    ]

                with m.Else():
                    m.d.sync += self.phy.cs.eq(0)


            # LATCH_RWDS -- latch in the value of the RWDS signal,
            # which determines our read/write latency.
            with m.State("LATCH_RWDS"):
                m.d.sync += extra_latency.eq(self.phy.rwds.i),
                m.d.sync += self.phy.clk_en.eq(0b11)
                m.next="SHIFT_COMMAND0"


            # SHIFT_COMMANDx -- shift each of our command words out
            with m.State('SHIFT_COMMAND0'):
                # Output the first 32 bits of our command.
                m.d.sync += [
                    self.phy.dq.o.eq(Cat(ca[16:48])),
                    self.phy.dq.e.eq(1),
                ]
                m.next = 'SHIFT_COMMAND1'

            with m.State('SHIFT_COMMAND1'):
                # Output the remaining 32 bits of our command.
                m.d.sync += [
                    self.phy.dq.o.eq(Cat(Const(0, 16), ca[0:16])),
                    self.phy.dq.e.eq(1),
                ]

                # If we have a register write, we don't need to handle
                # any latency. Move directly to our SHIFT_DATA state.
                with m.If(is_register & ~is_read):
                    m.next = 'WRITE_DATA'

                # Otherwise, react with either a short period of latency
                # or a longer one, depending on what the RAM requested via
                # RWDS.
                with m.Else():
                    m.next = "HANDLE_LATENCY"

                    # FIXME: our HyperRAM part has a fixed latency, but we could need to detect 
                    # different variants from the configuration register in the future.
                    with m.If(extra_latency | 1):
                        m.d.sync += latency_clocks_remaining.eq(self.HIGH_LATENCY_CLOCKS)
                    with m.Else():
                        m.d.sync += latency_clocks_remaining.eq(self.LOW_LATENCY_CLOCKS)


            # HANDLE_LATENCY -- applies clock cycles until our latency period is over.
            with m.State('HANDLE_LATENCY'):
                m.d.sync += latency_clocks_remaining.eq(latency_clocks_remaining - 1)

                with m.If(latency_clocks_remaining == 0):
                    with m.If(is_read):
                        m.next = 'READ_DATA'
                    with m.Else():
                        m.next = 'WRITE_DATA'


            # READ_DATA -- reads words from the PSRAM
            with m.State('READ_DATA'):
                m.d.sync += self.phy.read.eq(0b11)

                datavalid_delay = Signal()
                m.d.sync += datavalid_delay.eq(self.phy.datavalid)

                with m.If(self.phy.datavalid):
                    m.d.comb += [
                        self.read_data     .eq(self.phy.dq.i),
                        self.read_ready    .eq(1),
                    ]

                    # If our controller is done with the transaction, end it.
                    with m.If(self.final_word):
                        m.d.sync += self.phy.clk_en.eq(0),
                        m.next = 'RECOVERY'

            # WRITE_DATA -- write a word to the PSRAM
            with m.State("WRITE_DATA"):
                m.d.sync += [
                    self.phy.dq.o    .eq(self.write_data),
                    self.phy.dq.e    .eq(1),
                    self.phy.rwds.e  .eq(~is_register),
                    self.phy.rwds.o  .eq(0),
                ]
                m.d.comb += self.write_ready.eq(1),

                # If we just finished a register write, we're done -- there's no need for recovery.
                with m.If(is_register):
                    m.next = 'IDLE'

                with m.Elif(self.final_word):
                    m.next = 'RECOVERY'


            # RECOVERY state: wait for the required period of time before a new transaction
            with m.State('RECOVERY'):
                m.d.sync += self.phy.clk_en .eq(0)

                # TODO: implement recovery
                m.next = 'IDLE'



        return m


class HyperRAMDQSPHY(Elaboratable):
    """ Gateware PHY for HyperRAM series self-refreshing DRAM chips, using ECP5 DQS logic.

    I/O port:
        B: bus              -- The primary physical connection to the DRAM chip.
    """

    def __init__(self, *, bus, in_skew=None, out_skew=None, clock_skew=None):
        self.bus = bus
        self.phy = HyperBusDQSPHY()

    def elaborate(self, platform):
        m = Module()

        # Handle initial DDRDLL lock & delay code update
        pause = Signal()
        freeze = Signal()
        lock = Signal()
        uddcntln = Signal()
        counter = Signal(range(9))
        m.d.sync += counter.eq(counter + 1)
        with m.FSM() as fsm:
            with m.State('INIT'):
                m.d.sync += [
                    pause.eq(1),
                    freeze.eq(0),
                    uddcntln.eq(0),
                ]

                with m.If(lock):
                    m.next = 'FREEZE'
                    m.d.sync += [
                        freeze.eq(1),
                        counter.eq(0),
                    ]

            with m.State('FREEZE'):
                with m.If(counter == 8):
                    m.next = 'UPDATE'
                    m.d.sync += [
                        uddcntln.eq(1),
                        counter.eq(0),
                    ]

            with m.State('UPDATE'):
                with m.If(counter == 8):
                    m.next = 'UPDATED'
                    m.d.sync += [
                        uddcntln.eq(0),
                        counter.eq(0),
                    ]

            with m.State('UPDATED'):
                with m.If(counter == 8):
                    m.next = 'UNPAUSE'
                    m.d.sync += [
                        pause.eq(0),
                        counter.eq(0),
                    ]

            with m.State('UNPAUSE'):
                pass


        # DQS (RWDS) input
        rwds_o = Signal()
        rwds_oe_n = Signal()
        rwds_in = Signal()

        dqsr90 = Signal()
        dqsw = Signal()
        dqsw270 = Signal()
        ddrdel = Signal()
        readptr = Signal(3)
        writeptr = Signal(3)
        m.submodules += [
            Instance("DDRDLLA",
                i_CLK=ClockSignal("fast"),
                i_RST=ResetSignal(),
                i_FREEZE=freeze,
                i_UDDCNTLN=uddcntln,
                o_DDRDEL=ddrdel,
                o_LOCK=lock,
            ),
            Instance("BB",
                i_I=rwds_o,
                i_T=rwds_oe_n,
                o_O=rwds_in,
                io_B=self.bus.rwds.io
            ),
            Instance("TSHX2DQSA",
                i_RST=ResetSignal(),
                i_ECLK=ClockSignal("fast"),
                i_SCLK=ClockSignal(),
                i_DQSW=dqsw,
                i_T0=~self.phy.rwds.e,
                i_T1=~self.phy.rwds.e,
                o_Q=rwds_oe_n
            ),
            Instance("DQSBUFM",
                i_SCLK=ClockSignal(),
                i_ECLK=ClockSignal("fast"),
                i_RST=ResetSignal(),

                i_DQSI=rwds_in,
                i_DDRDEL=ddrdel,
                i_PAUSE=pause,
                i_READ0=self.phy.read[0],
                i_READ1=self.phy.read[1],
                # TODO: may need to tune at runtime by trying different values & checking for BURSTDET high
                i_READCLKSEL0=0,
                i_READCLKSEL1=1,
                i_READCLKSEL2=0,

                i_RDLOADN=0,
                i_RDMOVE=0,
                i_RDDIRECTION=1,
                i_WRLOADN=0,
                i_WRMOVE=0,
                i_WRDIRECTION=1,

                o_DQSR90=dqsr90,
                o_DQSW=dqsw,
                o_DQSW270=dqsw270,
                **{f"o_RDPNTR{i}": readptr[i] for i in range(len(readptr))},
                **{f"o_WRPNTR{i}": writeptr[i] for i in range(len(writeptr))},

                o_DATAVALID=self.phy.datavalid,
                o_BURSTDET=self.phy.burstdet,
            ),
        ]

        # Clock
        clk_out = Signal()
        clk_dqsw270 = Signal()
        m.submodules += [
            Instance("DELAYG",
                p_DEL_MODE="DQS_CMD_CLK",
                i_A=clk_out,
                o_Z=self.bus.clk,
            ),
            Instance("ODDRX2F",
                i_D0=0,
                i_D1=self.phy.clk_en[1],
                i_D2=0,
                i_D3=self.phy.clk_en[0],
                i_SCLK=ClockSignal(),
                i_ECLK=ClockSignal("fast"),
                i_RST=ResetSignal(),
                o_Q=clk_out,
            ),
        ]

        # CS
        cs_out = Signal()
        m.submodules += [
            Instance("DELAYG",
                p_DEL_MODE="DQS_CMD_CLK",
                i_A=cs_out,
                o_Z=self.bus.cs,
            ),
            Instance("ODDRX2F",
                i_D0=~self.phy.cs,
                i_D1=~self.phy.cs,
                i_D2=~self.phy.cs,
                i_D3=~self.phy.cs,
                i_SCLK=ClockSignal(),
                i_ECLK=ClockSignal("fast"),
                i_RST=ResetSignal(),
                o_Q=cs_out,
            ),
        ]

        # RWDS out
        m.submodules += [
            Instance("ODDRX2DQSB",
                i_DQSW=dqsw,
                i_D0=self.phy.rwds.o[3],
                i_D1=self.phy.rwds.o[2],
                i_D2=self.phy.rwds.o[1],
                i_D3=self.phy.rwds.o[0],
                i_SCLK=ClockSignal(),
                i_ECLK=ClockSignal("fast"),
                i_RST=ResetSignal(),
                o_Q=rwds_o,
            ),
        ]

        # DQ
        for i in range(8):
            dq_in   = Signal(name=f"dq_in{i}")
            dq_in_delayed   = Signal(name=f"dq_in_delayed{i}")
            dq_oe_n = Signal(name=f"dq_oe_n{i}")
            dq_o    = Signal(name=f"dq_o{i}")
            # Out
            m.submodules += [
                # Tristate
                Instance("BB",
                    i_I=dq_o,
                    i_T=dq_oe_n,
                    o_O=dq_in,
                    io_B=self.bus.dq.io[i]
                ),
                Instance("TSHX2DQA",
                    i_T0=~self.phy.dq.e,
                    i_T1=~self.phy.dq.e,
                    i_SCLK=ClockSignal(),
                    i_ECLK=ClockSignal("fast"),
                    i_DQSW270=dqsw270,
                    i_RST=ResetSignal(),
                    o_Q=dq_oe_n,
                ),

                # Output
                Instance("ODDRX2DQA",
                    i_DQSW270=dqsw270,
                    i_D0=self.phy.dq.o[i+24],
                    i_D1=self.phy.dq.o[i+16],
                    i_D2=self.phy.dq.o[i+8],
                    i_D3=self.phy.dq.o[i],
                    i_SCLK=ClockSignal(),
                    i_ECLK=ClockSignal("fast"),
                    i_RST=ResetSignal(),
                    o_Q=dq_o,
                ),

                # Input
                Instance("DELAYG",
                    p_DEL_MODE="DQS_ALIGNED_X2",
                    i_A=dq_in,
                    o_Z=dq_in_delayed,
                ),
                Instance("IDDRX2DQA",
                    i_D=dq_in_delayed,
                    i_DQSR90=dqsr90,
                    i_SCLK=ClockSignal(),
                    i_ECLK=ClockSignal("fast"),
                    i_RST=ResetSignal(),
                    **{f"i_RDPNTR{i}": readptr[i] for i in range(len(readptr))},
                    **{f"i_WRPNTR{i}": writeptr[i] for i in range(len(writeptr))},
                    o_Q0=self.phy.dq.i[i+24],
                    o_Q1=self.phy.dq.i[i+16],
                    o_Q2=self.phy.dq.i[i+8],
                    o_Q3=self.phy.dq.i[i],
                ),
            ]

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# SPDX-License-Identifier: BSD-3-Clause

""" Low-level USB transciever gateware -- control request components. """

import functools
import operator

from amaranth            import Signal, Module, Elaboratable, Cat
from amaranth.lib.coding import Encoder
from amaranth.lib.data   import Struct
from amaranth.hdl.rec    import Record, DIR_FANOUT

from .                   import USBSpeed
from .packet             import USBTokenDetector, USBDataPacketDeserializer
from .packet             import DataCRCInterface, USBInterpacketTimer, TokenDetectorInterface
from .packet             import InterpacketTimerInterface, HandshakeExchangeInterface
from ..stream            import USBInStreamInterface, USBOutStreamInterface
from ..request           import SetupPacket


class ClearEndpointHaltInterface(Struct):
    enable: 1
    direction: 1
    number: 4


class RequestHandlerInterface:
    """ Record representing a connection between a control endpoint and a request handler.

    Components (I = input to request handler; O = output to control interface):
        *: setup                  -- Carries the most recent setup request to the handler.
        *: tokenizer              -- Carries information about any incoming token packets.
        O: claim                  -- Assert to drive the rest of output signals.

        # Control request status signals.
        I: data_requested         -- Pulsed to indicate that a data-phase IN token has been issued,
                                     and it's now time to respond (post-inter-packet delay).
        I: status_requested       -- Pulsed to indicate that a response to our status phase has been
                                     requested.

        # Address / configuration connections.
        O: address_changed        -- Strobe; pulses high when the device's address should be changed.
        O: new_address[7]         -- When `address_changed` is high, this field contains the address that
                                     should be adopted.

        I: active_config          -- The configuration number of the active configuration.
        O: config_changed         -- Strobe; pulses high when the device's configuration should be changed.
        O: new_config[8]          -- When `config_changed` is high, this field contains the configuration that
                                     should be applied.

        # Data rx signals.
        *: rx                     -- The receive stream for any data packets received.
        I: handshakes_in          -- Inputs that indicate when handshakes are detected from the host.
        I: rx_ready_for_response  -- Strobe that indicates that we're ready to respond to a complete transmission.
                                     Indicates that an interpacket delay has passed after an `rx_complete` strobe.
        I: rx_invalid:            -- Strobe that indicates an invalid data receipt. Indicates that the most recently
                                     received packet was corrupted; and should be discarded as invalid.

        # Data tx signals.
        *: tx                     -- The transmit stream for any packets generated by the handler.
        O: handshakes_out         -- Carries handshake generation requests.
    """

    def __init__(self):
        self.setup                 = SetupPacket()
        self.tokenizer             = TokenDetectorInterface()
        self.claim                 = Signal()

        self.data_requested        = Signal()
        self.status_requested      = Signal()

        self.address_changed       = Signal()
        self.new_address           = Signal(7)

        self.active_config         = Signal(8)
        self.config_changed        = Signal()
        self.new_config            = Signal(8)

        self.clear_endpoint_halt   = Signal(ClearEndpointHaltInterface)

        self.rx                    = USBOutStreamInterface()
        self.rx_expected           = Signal()
        self.rx_ready_for_response = Signal()
        self.rx_invalid            = Signal()

        self.tx                    = USBInStreamInterface()
        self.handshakes_out        = HandshakeExchangeInterface(is_detector=True)
        self.handshakes_in         = HandshakeExchangeInterface(is_detector=False)
        self.tx_data_pid           = Signal(init=1)



class USBRequestHandler(Elaboratable):
    """ Base class for USB request handler modules.

    I/O port:
        *: interface -- The RequestHandlerInterface we'll use.
    """

    def __init__(self):

        #
        # I/O port:
        #
        self.interface = RequestHandlerInterface()


    def send_zlp(self):
        """ Returns the statements necessary to send a zero-length packet."""

        tx = self.interface.tx

        # Send a ZLP along our transmit interface.
        # Our interface accepts 'valid' and 'last' without 'first' as a ZLP.
        return [
            tx.valid  .eq(1),
            tx.last   .eq(1)
        ]





class USBSetupDecoder(Elaboratable):
    """ Gateware responsible for detecting Setup transactions.

    I/O port:
        *: data_crc  -- Interface to the device's data-CRC generator.
        *: tokenizer -- Interface to the device's token detector.
        *: timer     -- Interface to the device's interpacket timer.

        I: speed     -- The device's current operating speed. Should be a USBSpeed
                        enumeration value -- 0 for high, 1 for full, 2 for low.
        *: packet    -- The SetupPacket record carrying our parsed output.
        I: ack       -- True when we're requesting that an ACK be generated.
    """
    SETUP_PID = 0b1101

    def __init__(self, *, utmi, standalone=False):
        """
        Paremeters:
            utmi           -- The UTMI bus we'll monitor for data. We'll consider this read-only.

            standalone     -- Debug parameter. If true, this module will operate without external components;
                              i.e. without an internal data-CRC generator, or tokenizer. In this case, tokenizer
                              and timer should be set to None; and will be ignored.
        """
        self.utmi          = utmi
        self.standalone    = standalone

        #
        # I/O port.
        #
        self.data_crc      = DataCRCInterface()
        self.tokenizer     = TokenDetectorInterface()
        self.timer         = InterpacketTimerInterface()
        self.speed         = Signal(2)


        self.packet        = SetupPacket()
        self.ack           = Signal()


    def elaborate(self, platform):
        m = Module()

        # If we're standalone, generate the things we need.
        if self.standalone:

            # Create our tokenizer...
            m.submodules.tokenizer = tokenizer = USBTokenDetector(utmi=self.utmi)
            m.d.comb += tokenizer.interface.connect(self.tokenizer)

            # ... and our timer.
            m.submodules.timer = timer = USBInterpacketTimer()
            timer.add_interface(self.timer)

            m.d.comb += timer.speed.eq(self.speed)


        # Create a data-packet-deserializer, which we'll use to capture the
        # contents of the setup data packets.
        m.submodules.data_handler = data_handler = \
            USBDataPacketDeserializer(utmi=self.utmi, max_packet_size=8, create_crc_generator=self.standalone)
        m.d.comb += self.data_crc.connect(data_handler.data_crc)

        # Instruct our interpacket timer to begin counting when we complete receiving
        # our setup packet. This will allow us to track interpacket delays.
        m.d.comb += self.timer.start.eq(data_handler.new_packet)

        # Keep our output signals de-asserted unless specified.
        m.d.usb += [
            self.packet.received  .eq(0),
        ]


        with m.FSM(domain="usb"):

            # IDLE -- we haven't yet detected a SETUP transaction directed at us
            with m.State('IDLE'):
                pid_matches     = (self.tokenizer.pid     == self.SETUP_PID)

                # If we're just received a new SETUP token addressed to us,
                # the next data packet is going to be for us.
                with m.If(pid_matches & self.tokenizer.new_token):
                    m.next = 'READ_DATA'


            # READ_DATA -- we've just seen a SETUP token, and are waiting for the
            # data payload of the transaction, which contains the setup packet.
            with m.State('READ_DATA'):

                # If we receive a token packet before we receive a DATA packet,
                # this is a PID mismatch. Bail out and start over.
                with m.If(self.tokenizer.new_token):
                    m.next = 'IDLE'

                # If we have a new packet, parse it as setup data.
                with m.If(data_handler.new_packet):

                    # If we got exactly eight bytes, this is a valid setup packet.
                    with m.If(data_handler.length == 8):

                        # Collect the signals that make up our bmRequestType [USB2, 9.3].
                        request_type = Cat(self.packet.recipient, self.packet.type, self.packet.is_in_request)

                        m.d.usb += [

                            # Parse the setup data itself...
                            request_type            .eq(data_handler.packet[0]),
                            self.packet.request     .eq(data_handler.packet[1]),
                            self.packet.value       .eq(Cat(data_handler.packet[2], data_handler.packet[3])),
                            self.packet.index       .eq(Cat(data_handler.packet[4], data_handler.packet[5])),
                            self.packet.length      .eq(Cat(data_handler.packet[6], data_handler.packet[7])),

                            # ... and indicate that we have new data.
                            self.packet.received  .eq(1),

                        ]

                        # We'll now need to wait a receive-transmit delay before initiating our ACK.
                        # Per the USB 2.0 and ULPI 1.1 specifications:
                        #   - A HS device needs to wait 8 HS bit periods before transmitting [USB2, 7.1.18.2].
                        #     Each ULPI cycle is 8 HS bit periods, so we'll only need to wait one cycle.
                        #   - We'll use our interpacket delay timer for everything else.
                        with m.If(self.timer.tx_allowed | (self.speed == USBSpeed.HIGH)):

                            # If we're a high speed device, we only need to wait for a single ULPI cycle.
                            # Processing delays mean we've already met our interpacket delay; and we can ACK
                            # immediately.
                            m.d.comb += self.ack.eq(1)
                            m.next = "IDLE"

                        # For other cases, handle the interpacket delay by waiting.
                        with m.Else():
                            m.next = "INTERPACKET_DELAY"


                    # Otherwise, this isn't; and we should ignore it. [USB2, 8.5.3]
                    with m.Else():
                        m.next = "IDLE"


            # INTERPACKET -- wait for an inter-packet delay before responding
            with m.State('INTERPACKET_DELAY'):

                # ... and once it equals zero, ACK and return to idle.
                with m.If(self.timer.tx_allowed):
                    m.d.comb += self.ack.eq(1)
                    m.next = "IDLE"

        return m



class USBRequestHandlerMultiplexer(Elaboratable):
    """ Multiplexes multiple RequestHandlers down to a single interface.

    Interfaces are added using .add_interface().

    I/O port:
        *: shared -- The post-multiplexer RequestHandler interface.
    """

    def __init__(self):

        #
        # I/O port
        #
        self.shared = RequestHandlerInterface()

        #
        # Internals
        #
        self._interfaces = []
        self._fallback   = None


    def add_interface(self, interface: RequestHandlerInterface):
        """ Adds a RequestHandlerInterface to the multiplexer.

        It's expected only one handler will be driving requests at a time.
        """
        self._interfaces.append(interface)

    def set_fallback_interface(self, interface: RequestHandlerInterface):
        """ Sets a RequestHandlerInterface as a fallback for unhandled requests. """
        self._fallback = interface

    def _multiplex_signals(self, m, *, when, multiplex, sub_bus=None):
        """ Helper that creates a simple priority-encoder multiplexer.

        Parmeters:
            when      -- The name of the interface signal that indicates that the `multiplex` signals
                         should be selected for output. If this signals should be multiplex, it
                         should be included in `multiplex`.
            multiplex -- The names of the interface signals to be multiplexed.
        """

        def get_signal(interface, name):
            """ Fetches an interface signal by name / sub_bus. """

            if sub_bus:
                bus = getattr(interface, sub_bus)
                return getattr(bus, name)
            else:
                return  getattr(interface, name)


        # We're building an if-elif tree; so we should start with an If entry.
        conditional = m.If

        for interface in self._interfaces:
            condition = get_signal(interface, when)

            with conditional(condition):

                # Connect up each of our signals.
                for signal_name in multiplex:

                    # Get the actual signals for our input and output...
                    driving_signal = get_signal(interface,   signal_name)
                    target_signal  = get_signal(self.shared, signal_name)

                    # ... and connect them.
                    m.d.comb += target_signal   .eq(driving_signal)

            # After the first element, all other entries should be created with Elif.
            conditional = m.Elif



    def elaborate(self, platform):
        m = Module()
        shared = self.shared

        # If no fallback request handler is provided, stall all unhandled requests.
        if self._fallback is None:
            m.submodules.stall_handler = stall_handler = StallOnlyRequestHandler()
            self._fallback = stall_handler.interface

        #
        # Pass through signals being routed -to- our pre-mux interfaces.
        #
        for interface in [*self._interfaces, self._fallback]:
            m.d.comb += [
                shared.setup                     .connect(interface.setup),
                shared.tokenizer                 .connect(interface.tokenizer),

                interface.data_requested         .eq(shared.data_requested),
                interface.status_requested       .eq(shared.status_requested),
                shared.handshakes_in             .connect(interface.handshakes_in),
                interface.active_config          .eq(shared.active_config),

                shared.rx                        .connect(interface.rx),
                interface.rx_ready_for_response  .eq(shared.rx_ready_for_response),
                interface.rx_invalid             .eq(shared.rx_invalid),
            ]

        #
        # Multiplex the signals being routed -from- our pre-mux interface.
        #

        def _connect_interface_outputs(interface):
            m.d.comb += [
                shared.tx                 .stream_eq(interface.tx),

                shared.tx_data_pid        .eq(interface.tx_data_pid),

                shared.handshakes_out     .eq(interface.handshakes_out),

                shared.address_changed    .eq(interface.address_changed),
                shared.new_address        .eq(interface.new_address),
                shared.config_changed     .eq(interface.config_changed),
                shared.new_config         .eq(interface.new_config),
                shared.clear_endpoint_halt.eq(interface.clear_endpoint_halt),
            ]

        # The encoder provides the index of the single interface that claims the 
        # output lines. Otherwise, it asserts the .n (invalid) line.
        m.submodules.encoder = encoder = Encoder(len(self._interfaces))
        m.d.comb += encoder.i.eq(Cat(interface.claim for interface in self._interfaces))

        # Connect the interface outputs to the interface that claims them.
        with m.Switch(encoder.o):
            for index, interface in enumerate(self._interfaces):
                with m.Case(index):
                    _connect_interface_outputs(interface)

        # Use the fallback handler interface for the invalid case.
        with m.If(encoder.n):
            _connect_interface_outputs(self._fallback)
        
        return m


class StallOnlyRequestHandler(Elaboratable):
    """ Simple gateware request handler that only conditionally stalls requests.

    I/O port:
        *: interface -- The RequestHandlerInterface used to handle requests.
                        See its record definition for signal definitions.
    """

    def __init__(self, stall_condition=None):
        """
        Parameters:
            stall_condition -- A function that accepts a SetupRequest packet, and returns
                               an Amaranth conditional indicating whether we should stall.
        """

        self.condition = stall_condition or (lambda _: 1)

        #
        # I/O port
        #
        self.interface = RequestHandlerInterface()


    def elaborate(self, platform):
        m = Module()

        # If we have an opportunity to stall...
        with m.If(self.interface.data_requested | self.interface.status_requested):

            # ... and our stall condition is met ...
            with m.If(self.condition(self.interface.setup)):

                # ... do so.
                m.d.comb += self.interface.handshakes_out.stall.eq(1)

        return m



#
# This file is part of LUNA.
#
# Copyright (c) 2020 Great Scott Gadgets <info@greatscottgadgets.com>
# Copyright (c) 2020 Florent Kermarrec <florent@enjoy-digital.fr>
#
# Code based in part on ``litex`` and ``liteiclink``.
# SPDX-License-Identifier: BSD-3-Clause
""" Soft PIPE backend for the Xilinx 7 Series GTX transceivers. """

from amaranth import *
from amaranth.lib.cdc import FFSynchronizer


from .xc7           import DRPInterface, DRPArbiter, DRPFieldController
from .xc7           import GTResetDeferrer, GTOOBClockDivider
from .lfps         import LFPSSquareWaveGenerator, LFPSSquareWaveDetector
from ..pipe        import PIPEInterface


Open = Signal


class GTXQuadPLL(Elaboratable):
    def __init__(self, refclk, refclk_freq, linerate):
        self._refclk      = refclk
        self._refclk_freq = refclk_freq
        self._linerate    = linerate

        self.config  = self.compute_config(refclk_freq, linerate)

        #
        # I/O ports
        #
        self.clk     = Signal()
        self.refclk  = Signal()
        self.reset   = Signal()
        self.lock    = Signal()
        self.drp     = DRPInterface()


    def elaborate(self, platform):
        qpll_fbdivs = {
            16:  0b0000100000,
            20:  0b0000110000,
            32:  0b0001100000,
            40:  0b0010000000,
            64:  0b0011100000,
            66:  0b0101000000,
            80:  0b0100100000,
            100: 0b0101110000,
        }

        qpll_fbdiv_ratios = {
            16:  1,
            20:  1,
            32:  1,
            40:  1,
            64:  1,
            66:  0,
            80:  1,
            100: 1,
        }

        return Instance("GTXE2_COMMON",
            # Common Block Attributes
            p_BIAS_CFG                 = 0x0000040000001000,
            p_COMMON_CFG               = 0x00000000,
            p_QPLL_CFG                 = 0x0680181 if self.config["vco_band"] == "upper" else 0x06801C1,
            p_QPLL_CLKOUT_CFG          = 0b0000,
            p_QPLL_COARSE_FREQ_OVRD    = 0b010000,
            p_QPLL_COARSE_FREQ_OVRD_EN = 0b0,
            p_QPLL_CP                  = 0b0000011111,
            p_QPLL_CP_MONITOR_EN       = 0b0,
            p_QPLL_DMONITOR_SEL        = 0b0,
            p_QPLL_FBDIV               = qpll_fbdivs[self.config["n"]],
            p_QPLL_FBDIV_MONITOR_EN    = 0b0,
            p_QPLL_FBDIV_RATIO         = qpll_fbdiv_ratios[self.config["n"]],
            p_QPLL_INIT_CFG            = 0x000006,
            p_QPLL_LOCK_CFG            = 0x21E8,
            p_QPLL_LPF                 = 0b1111,
            p_QPLL_REFCLK_DIV          = self.config["m"],

            # Common Block - Dynamic Reconfiguration Port
            i_DRPCLK            = ClockSignal("ss"),
            i_DRPADDR           = self.drp.addr,
            i_DRPDI             = self.drp.di,
            o_DRPDO             = self.drp.do,
            i_DRPWE             = self.drp.we,
            i_DRPEN             = self.drp.en,
            o_DRPRDY            = self.drp.rdy,

            # Common Block - Ref Clock Ports
            i_GTGREFCLK         = 0,
            i_GTNORTHREFCLK0    = 0,
            i_GTNORTHREFCLK1    = 0,
            i_GTREFCLK0         = self._refclk,
            i_GTREFCLK1         = 0,
            i_GTSOUTHREFCLK0    = 0,
            i_GTSOUTHREFCLK1    = 0,

            # Common Block - QPLL Ports
            o_QPLLDMONITOR      = Open(8),

            # Common Block - Clocking Ports
            o_QPLLOUTCLK        = self.clk,
            o_QPLLOUTREFCLK     = self.refclk,
            o_REFCLKOUTMONITOR  = Open(),

            # Common Block - QPLL Ports
            i_BGBYPASSB         = 1,
            i_BGMONITORENB      = 1,
            i_BGPDB             = 1,
            i_BGRCALOVRD        = 0b11111,
            i_PMARSVD           = 0b00000000,
            i_RCALENB           = 1,
            o_QPLLFBCLKLOST     = Open(),
            o_QPLLLOCK          = self.lock,
            i_QPLLLOCKDETCLK    = 0,
            i_QPLLLOCKEN        = 1,
            i_QPLLOUTRESET      = 0,
            i_QPLLPD            = 0,
            o_QPLLREFCLKLOST    = Open(),
            i_QPLLREFCLKSEL     = 0b001,
            i_QPLLRESET         = self.reset,
            i_QPLLRSVD1         = 0b0000000000000000,
            i_QPLLRSVD2         = 0b11111,
        )


    @staticmethod
    def compute_config(refclk_freq, linerate):
        for m in 1, 2, 3, 4:
            for n in 16, 20, 32, 40, 64, 66, 80, 100:
                vco_freq = refclk_freq*n/m
                if 5.93e9 <= vco_freq <= 8e9:
                    vco_band = "lower"
                elif 9.8e9 <= vco_freq <= 12.5e9:
                    vco_band = "upper"
                else:
                    vco_band = None
                if vco_band is not None:
                    for d in 1, 2, 4, 8, 16:
                        current_linerate = (vco_freq/2)*2/d
                        if current_linerate == linerate:
                            return {"n": n, "nr": 0 if n == 66 else 1, "m": m, "d": d,
                                    "vco_freq": vco_freq,
                                    "vco_band": vco_band,
                                    "clkin": refclk_freq,
                                    "linerate": linerate}
        msg = "No config found for {:3.2f} MHz refclk / {:3.2f} Gbps linerate."
        raise ValueError(msg.format(refclk_freq/1e6, linerate/1e9))

    def __repr__(self):
        config = self.config
        r = """
GTXQuadPLL
==========
  overview:
  ---------
       +---------------------------------------------------------------+
       |                                                               |
       |           +---------------------------+ +------------+        |
       |  +-----+  | Phase Frequency Detector  | | Upper Band | +----+ |
CLKIN +---> /M  +-->       Charge Pump         +->    VCO     +-> /2 +--> CLKOUT
       |  +-----+  |       Loop Filter         | | Lower Band | +----+ |
       |           +---------------------------+ +-----+------+        |
       |             ^                                 |               |
       |             |    +-----+                      |               |
       |             +----+ /N  <----------------------+               |
       |                  +-----+                                      |
       +---------------------------------------------------------------+
                            +-------+
                   CLKOUT +->  2/D  +-> LINERATE
                            +-------+
  config:
  -------
    CLKIN    = {clkin}MHz
    CLKOUT   = CLKIN x N / (2 x M) = {clkin}MHz x {n} / (2 x {m})
             = {vco_freq}GHz
    LINERATE = CLKOUT x 2 / D = {vco_freq}GHz x 2 / {d}
             = {linerate}GHz
""".format(clkin    = config["clkin"]/1e6,
           n        = config["n"],
           m        = config["m"],
           vco_freq = config["vco_freq"]/1e9,
           d        = config["d"],
           linerate = config["linerate"]/1e9)
        return r


class GTXChannelPLL(Elaboratable):
    def __init__(self, refclk, refclk_freq, linerate):
        self._refclk      = refclk
        self._refclk_freq = refclk_freq
        self._linerate    = linerate

        self.config  = self.compute_config(refclk_freq, linerate)

        #
        # I/O ports
        #
        self.reset   = Signal()
        self.lock    = Signal()


    def elaborate(self, platform):
        return Module()


    @staticmethod
    def compute_config(refclk_freq, linerate):
        for m in 1, 2:
            for n2 in 1, 2, 3, 4, 5:
                for n1 in 4, 5:
                    vco_freq = refclk_freq*(n1*n2)/m
                    if 1.6e9 <= vco_freq <= 3.3e9:
                        for d in 1, 2, 4, 8, 16:
                            current_linerate = vco_freq*2/d
                            if current_linerate == linerate:
                                return {"n1": n1, "n2": n2, "m": m, "d": d,
                                        "vco_freq": vco_freq,
                                        "clkin": refclk_freq,
                                        "linerate": linerate}
        msg = "No config found for {:3.2f} MHz refclk / {:3.2f} Gbps linerate."
        raise ValueError(msg.format(refclk_freq/1e6, linerate/1e9))

    def __repr__(self):
        config = self.config
        r = """
GTXChannelPLL
=============
  overview:
  ---------
       +--------------------------------------------------+
       |                                                  |
       |            +---------------------------+ +-----+ |
       |   +-----+  | Phase Frequency Detector  | |     | |
CLKIN +----> /M  +-->       Charge Pump         +-> VCO +---> CLKOUT
       |   +-----+  |       Loop Filter         | |     | |
       |            +---------------------------+ +--+--+ |
       |              ^                              |    |
       |              |    +-------+    +-------+    |    |
       |              +----+  /N2  <----+  /N1  <----+    |
       |                   +-------+    +-------+         |
       +--------------------------------------------------+
                            +-------+
                   CLKOUT +->  2/D  +-> LINERATE
                            +-------+
  config:
  -------
    CLKIN    = {clkin}MHz
    CLKOUT   = CLKIN x (N1 x N2) / M = {clkin}MHz x ({n1} x {n2}) / {m}
             = {vco_freq}GHz
    LINERATE = CLKOUT x 2 / D = {vco_freq}GHz x 2 / {d}
             = {linerate}GHz
""".format(clkin    = config["clkin"]/1e6,
           n1       = config["n1"],
           n2       = config["n2"],
           m        = config["m"],
           vco_freq = config["vco_freq"]/1e9,
           d        = config["d"],
           linerate = config["linerate"]/1e9)
        return r


class GTXChannel(Elaboratable):
    def __init__(self, pll, tx_pads, rx_pads, ss_clock_frequency):
        self._pll           = pll
        self._tx_pads       = tx_pads
        self._rx_pads       = rx_pads
        self._ss_clock_frequency = ss_clock_frequency

        # For now, always operate at 2x gearing, and using the corresponding width for
        # the internal data path.
        self._io_words      = 2
        self._data_width    = self._io_words * 10

        #
        # I/O ports.
        #

        # Dynamic reconfiguration port
        self.drp            = DRPInterface()

        # Interface clock
        self.pclk           = Signal()

        # Reset sequencing
        self.reset          = Signal()
        self.tx_ready       = Signal()
        self.rx_ready       = Signal()

        # Core Rx and Tx lines
        self.tx_data        = Signal(self._io_words * 8)
        self.tx_datak       = Signal(self._io_words)
        self.rx_data        = Signal(self._io_words * 8)
        self.rx_datak       = Signal(self._io_words)

        # TX controls
        self.tx_polarity    = Signal()
        self.tx_elec_idle   = Signal()
        self.tx_gpio_en     = Signal()
        self.tx_gpio        = Signal()

        # RX controls
        self.rx_polarity    = Signal()
        self.rx_eq_training = Signal()
        self.rx_termination = Signal()

        # RX status
        self.rx_valid       = Signal()
        self.rx_status      = Signal(3)
        self.rx_elec_idle   = Signal()


    def elaborate(self, platform):
        m = Module()

        # Aliases.
        pll        = self._pll
        io_words   = self._io_words
        data_width = self._data_width

        #
        # Clocking.
        #

        # Ensure we have a valid PLL/CDR configuration.
        assert pll.config["linerate"] < 6.6e9

        # We support both QPLL and CPLL clock sources.
        use_cpll = isinstance(pll, GTXChannelPLL)
        use_qpll = isinstance(pll, GTXQuadPLL)

        # From [UG476: Table 4-18]: GTX CDR Recommended Settings for Protocols with SSC
        rxcdr_cfgs = {
            1: 0x03_8000_8BFF_1020_0010,
            2: 0x03_8000_8BFF_4020_0008,
            4: 0x03_8000_8BFF_4010_0008,
        }

        # Generate the PIPE interface clock from the transmit word clock, and use it to drive both
        # the Tx and the Rx FIFOs, to bring both halves of the data bus to the same clock domain.
        # The recovered Rx clock will not match the generated Tx clock; use the recovered word
        # clock to drive the CTC FIFO in the transceiver, which will compensate for the difference.
        txoutclk = Signal()
        m.submodules += Instance("BUFG",
            i_I=txoutclk,
            o_O=self.pclk
        )
        platform.add_clock_constraint(self.pclk, 250e6)

        # Transceiver uses a 25 MHz clock internally, which needs to be derived from
        # the reference clock.
        for clk25_div in range(1, 33):
            if pll._refclk_freq / clk25_div <= 25e6:
                break

        # Out of band sequence detector uses an auxiliary clock whose frequency is derived
        # from the properties of the sequences.
        m.submodules.oob_clkdiv = oob_clkdiv = GTOOBClockDivider(self._ss_clock_frequency)


        #
        # Initialization.
        #

        # Per [AR43482], GTX transceivers must not be reset immediately after configuration.
        m.submodules.defer_rst = defer_rst = GTResetDeferrer(self._ss_clock_frequency)
        m.d.comb += [
            defer_rst.tx_i.eq(~pll.lock | self.reset),
            defer_rst.rx_i.eq(~pll.lock | self.reset),
        ]

        tx_rst_done = Signal()
        rx_rst_done = Signal()
        m.d.comb += [
            self.tx_ready.eq(defer_rst.done & tx_rst_done),
            self.rx_ready.eq(defer_rst.done & rx_rst_done),
        ]


        #
        # Dynamic reconfiguration.
        #
        rx_termination = Signal()
        m.submodules += FFSynchronizer(self.rx_termination, rx_termination, o_domain="ss")

        m.submodules.rx_term = rx_term = DRPFieldController(
            addr=0x0011, bits=slice(4, 6), init=0b10) # RX_CM_SEL
        m.d.comb += [
            rx_term.value.eq(Mux(self.rx_termination,
                                 0b11,    # Programmable
                                 0b10)),  # Floating
        ]

        m.submodules.drp_arbiter = drp_arbiter = DRPArbiter()
        drp_arbiter.add_interface(rx_term.drp)
        drp_arbiter.add_interface(self.drp)


        #
        # Core SerDes instantiation.
        #
        m.submodules.gtx = Instance("GTXE2_CHANNEL",
            # Simulation-Only Attributes
            p_SIM_RECEIVER_DETECT_PASS   = "TRUE",
            p_SIM_TX_EIDLE_DRIVE_LEVEL   = "X",
            p_SIM_RESET_SPEEDUP          = "FALSE",
            p_SIM_CPLLREFCLK_SEL         = 0b001,
            p_SIM_VERSION                = "4.0",

            # RX 8B/10B Decoder Attributes
            p_RX_DISPERR_SEQ_MATCH       = "FALSE",
            p_DEC_MCOMMA_DETECT          = "TRUE",
            p_DEC_PCOMMA_DETECT          = "TRUE",
            p_DEC_VALID_COMMA_ONLY       = "TRUE",
            p_UCODEER_CLR                = 0b0,

            # RX Byte and Word Alignment Attributes
            p_ALIGN_COMMA_DOUBLE         = "FALSE",
            p_ALIGN_COMMA_ENABLE         = 0b1111_111111,
            p_ALIGN_COMMA_WORD           = 1,
            p_ALIGN_MCOMMA_DET           = "TRUE",
            p_ALIGN_MCOMMA_VALUE         = 0b0101_111100, # K28.5 RD- 10b code
            p_ALIGN_PCOMMA_DET           = "TRUE",
            p_ALIGN_PCOMMA_VALUE         = 0b1010_000011, # K28.5 RD+ 10b code
            p_SHOW_REALIGN_COMMA         = "TRUE",
            p_RXSLIDE_AUTO_WAIT          = 7,
            p_RXSLIDE_MODE               = "OFF",
            p_RX_SIG_VALID_DLY           = 10,

            # RX Clock Correction Attributes
            p_CBCC_DATA_SOURCE_SEL       = "DECODED",
            p_CLK_CORRECT_USE            = "TRUE",
            p_CLK_COR_KEEP_IDLE          = "FALSE",
            p_CLK_COR_MAX_LAT            = 14,
            p_CLK_COR_MIN_LAT            = 11,
            p_CLK_COR_PRECEDENCE         = "TRUE",
            p_CLK_COR_REPEAT_WAIT        = 0,
            p_CLK_COR_SEQ_LEN            = 2,
            p_CLK_COR_SEQ_1_ENABLE       = 0b1111,
            p_CLK_COR_SEQ_1_1            = 0b01_001_11100, # K28.1 1+8b code
            p_CLK_COR_SEQ_1_2            = 0b01_001_11100, # K28.1 1+8b code
            p_CLK_COR_SEQ_1_3            = 0b0000000000,
            p_CLK_COR_SEQ_1_4            = 0b0000000000,
            p_CLK_COR_SEQ_2_ENABLE       = 0b1111,
            p_CLK_COR_SEQ_2_USE          = "FALSE",
            p_CLK_COR_SEQ_2_1            = 0b0000000000,
            p_CLK_COR_SEQ_2_2            = 0b0000000000,
            p_CLK_COR_SEQ_2_3            = 0b0000000000,
            p_CLK_COR_SEQ_2_4            = 0b0000000000,

            # RX Channel Bonding Attributes
            p_CHAN_BOND_KEEP_ALIGN       = "FALSE",
            p_CHAN_BOND_MAX_SKEW         = 1,
            p_CHAN_BOND_SEQ_LEN          = 1,
            p_CHAN_BOND_SEQ_1_1          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_2          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_3          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_4          = 0b0000000000,
            p_CHAN_BOND_SEQ_1_ENABLE     = 0b1111,
            p_CHAN_BOND_SEQ_2_1          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_2          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_3          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_4          = 0b0000000000,
            p_CHAN_BOND_SEQ_2_ENABLE     = 0b1111,
            p_CHAN_BOND_SEQ_2_USE        = "FALSE",
            p_FTS_DESKEW_SEQ_ENABLE      = 0b1111,
            p_FTS_LANE_DESKEW_CFG        = 0b1111,
            p_FTS_LANE_DESKEW_EN         = "FALSE",

            # RX Margin Analysis Attributes
            p_ES_CONTROL                 = 0b000000,
            p_ES_ERRDET_EN               = "FALSE",
            p_ES_EYE_SCAN_EN             = "TRUE",
            p_ES_HORZ_OFFSET             = 0x000,
            p_ES_PMA_CFG                 = 0b0000000000,
            p_ES_PRESCALE                = 0b00000,
            p_ES_QUALIFIER               = 0x00000000000000000000,
            p_ES_QUAL_MASK               = 0x00000000000000000000,
            p_ES_SDATA_MASK              = 0x00000000000000000000,
            p_ES_VERT_OFFSET             = 0b000000000,

            # FPGA RX Interface Attributes
            p_RX_DATA_WIDTH              = data_width,

            # PMA Attributes
            p_DMONITOR_CFG               = 0x000A00,
            p_OUTREFCLK_SEL_INV          = 0b11,
            p_PMA_RSV                    = 0x00018480,
            p_PMA_RSV2                   = 0x00002050,
            p_PMA_RSV3                   = 0b00,
            p_PMA_RSV4                   = 0x00000000,
            p_RX_BIAS_CFG                = 0b000000000100,
            p_RX_CM_SEL                  = 0b10,
            p_RX_CM_TRIM                 = 0b010,
            p_RX_DEBUG_CFG               = 0b000000000000,
            p_RX_OS_CFG                  = 0b0000010000000,
            p_TERM_RCAL_CFG              = 0b10000,
            p_TERM_RCAL_OVRD             = 0b0,
            p_TST_RSV                    = 0x00000000,
            p_RX_CLK25_DIV               = clk25_div,
            p_TX_CLK25_DIV               = clk25_div,

            # PCI Express Attributes
            p_PCS_PCIE_EN                = "FALSE",

            # PCS Attributes
            p_PCS_RSVD_ATTR              = 0x0000_0000_0108, # OOB power up, OOB clock from CLKRSVD[0]

            # RX Buffer Attributes
            p_RXBUF_ADDR_MODE            = "FULL",
            p_RXBUF_EIDLE_HI_CNT         = 0b1000,
            p_RXBUF_EIDLE_LO_CNT         = 0b0000,
            p_RXBUF_EN                   = "TRUE",
            p_RX_BUFFER_CFG              = 0b000000,
            p_RXBUF_RESET_ON_CB_CHANGE   = "TRUE",
            p_RXBUF_RESET_ON_COMMAALIGN  = "FALSE",
            p_RXBUF_RESET_ON_EIDLE       = "FALSE",
            p_RXBUF_RESET_ON_RATE_CHANGE = "TRUE",
            p_RXBUFRESET_TIME            = 0b00001,
            p_RXBUF_THRESH_OVFLW         = 61,
            p_RXBUF_THRESH_OVRD          = "FALSE",
            p_RXBUF_THRESH_UNDFLW        = 4,
            p_RXDLY_CFG                  = 0x001F,
            p_RXDLY_LCFG                 = 0x030,
            p_RXDLY_TAP_CFG              = 0x0000,
            p_RXPH_CFG                   = 0x000000,
            p_RXPHDLY_CFG                = 0x084020,
            p_RXPH_MONITOR_SEL           = 0b00000,
            p_RX_XCLK_SEL                = "RXREC",
            p_RX_DDI_SEL                 = 0b000000,
            p_RX_DEFER_RESET_BUF_EN      = "TRUE",

            # CDR Attributes
            p_RXCDR_CFG                  = rxcdr_cfgs[pll.config["d"]],
            p_RXCDR_FR_RESET_ON_EIDLE    = 0b0,
            p_RXCDR_HOLD_DURING_EIDLE    = 0b0,
            p_RXCDR_PH_RESET_ON_EIDLE    = 0b0,
            p_RXCDR_LOCK_CFG             = 0b010101,

            # RX Initialization and Reset Attributes
            p_RXCDRFREQRESET_TIME        = 0b00001,
            p_RXCDRPHRESET_TIME          = 0b00001,
            p_RXDFELPMRESET_TIME         = 0b0001111,
            p_RXISCANRESET_TIME          = 0b00001,
            p_RXPCSRESET_TIME            = 0b00001,
            p_RXPMARESET_TIME            = 0b00011,

            # RX OOB Signaling Attributes
            p_RXOOB_CFG                  = 0b0000110,

            # RX Gearbox Attributes
            p_RXGEARBOX_EN               = "FALSE",
            p_GEARBOX_MODE               = 0b000,

            # PRBS Detection Attribute
            p_RXPRBS_ERR_LOOPBACK        = 0b0,

            # Power-Down Attributes
            p_PD_TRANS_TIME_FROM_P2      = 0x03c,
            p_PD_TRANS_TIME_NONE_P2      = 0x3c,
            p_PD_TRANS_TIME_TO_P2        = 0x64,

            # RX OOB Signaling Attributes
            p_SAS_MAX_COM                = 64,
            p_SAS_MIN_COM                = 36,
            p_SATA_BURST_SEQ_LEN         = 0b0101,
            p_SATA_BURST_VAL             = 0b100,
            p_SATA_EIDLE_VAL             = 0b100,
            p_SATA_MAX_BURST             = 8,
            p_SATA_MAX_INIT              = 21,
            p_SATA_MAX_WAKE              = 7,
            p_SATA_MIN_BURST             = 4,
            p_SATA_MIN_INIT              = 12,
            p_SATA_MIN_WAKE              = 4,

            # RX Fabric Clock Output Control Attributes
            p_TRANS_TIME_RATE            = 0x0E,

            # TX Buffer Attributes
            p_TXBUF_EN                   = "TRUE",
            p_TXBUF_RESET_ON_RATE_CHANGE = "TRUE",
            p_TXDLY_CFG                  = 0x001F,
            p_TXDLY_LCFG                 = 0x030,
            p_TXDLY_TAP_CFG              = 0x0000,
            p_TXPH_CFG                   = 0x0780,
            p_TXPHDLY_CFG                = 0x084020,
            p_TXPH_MONITOR_SEL           = 0b00000,
            p_TX_XCLK_SEL                = "TXOUT",

            # FPGA TX Interface Attributes
            p_TX_DATA_WIDTH              = data_width,

            # TX Configurable Driver Attributes
            p_TX_DEEMPH0                 = 0b000000,
            p_TX_DEEMPH1                 = 0b000000,
            p_TX_DRIVE_MODE              = "DIRECT",
            p_TX_EIDLE_ASSERT_DELAY      = 0b110,
            p_TX_EIDLE_DEASSERT_DELAY    = 0b100,
            p_TX_LOOPBACK_DRIVE_HIZ      = "FALSE",
            p_TX_MAINCURSOR_SEL          = 0b0,
            p_TX_PREDRIVER_MODE          = 0b0,
            p_TX_MARGIN_FULL_0           = 0b1001110,
            p_TX_MARGIN_FULL_1           = 0b1001001,
            p_TX_MARGIN_FULL_2           = 0b1000101,
            p_TX_MARGIN_FULL_3           = 0b1000010,
            p_TX_MARGIN_FULL_4           = 0b1000000,
            p_TX_MARGIN_LOW_0            = 0b1000110,
            p_TX_MARGIN_LOW_1            = 0b1000100,
            p_TX_MARGIN_LOW_2            = 0b1000010,
            p_TX_MARGIN_LOW_3            = 0b1000000,
            p_TX_MARGIN_LOW_4            = 0b1000000,
            p_TX_QPI_STATUS_EN           = 0b0,

            # TX Gearbox Attributes
            p_TXGEARBOX_EN               = "FALSE",

            # TX Initialization and Reset Attributes
            p_TXPCSRESET_TIME            = 0b00001,
            p_TXPMARESET_TIME            = 0b00001,

            # TX Receiver Detection Attributes
            p_TX_RXDETECT_CFG            = 0x1832,
            p_TX_RXDETECT_REF            = 0b100,

            # CPLL Attributes
            p_CPLL_CFG                   = 0xBC07DC,
            p_CPLL_FBDIV                 = pll.config["n2"] if use_cpll else 1,
            p_CPLL_FBDIV_45              = pll.config["n1"] if use_cpll else 4,
            p_CPLL_INIT_CFG              = 0x00001E,
            p_CPLL_LOCK_CFG              = 0x01E8,
            p_CPLL_REFCLK_DIV            = pll.config["m"]  if use_cpll else 1,
            p_RXOUT_DIV                  = pll.config["d"],
            p_TXOUT_DIV                  = pll.config["d"],
            p_SATA_CPLL_CFG              = "VCO_3000MHZ",

            # RX Equalizer Attributes
            p_RXLPM_HF_CFG               = 0b00000011110000,
            p_RXLPM_LF_CFG               = 0b00000011110000,
            p_RX_DFE_GAIN_CFG            = 0x020FEA,
            p_RX_DFE_H2_CFG              = 0b000000000000,
            p_RX_DFE_H3_CFG              = 0b000001000000,
            p_RX_DFE_H4_CFG              = 0b00011110000,
            p_RX_DFE_H5_CFG              = 0b00011100000,
            p_RX_DFE_KL_CFG              = 0b0000011111110,
            p_RX_DFE_KL_CFG2             = 0x301148AC,
            p_RX_DFE_LPM_CFG             = 0x0904,
            p_RX_DFE_LPM_HOLD_DURING_EIDLE = 0b0,
            p_RX_DFE_UT_CFG              = 0b10001111000000000,
            p_RX_DFE_VP_CFG              = 0b00011111100000011,
            p_RX_DFE_XYD_CFG             = 0b0000000000000,

            # Power-Down Attributes
            p_RX_CLKMUX_PD               = 0b1,
            p_TX_CLKMUX_PD               = 0b1,

            # FPGA RX Interface Attribute
            p_RX_INT_DATAWIDTH           = 0,

            # FPGA TX Interface Attribute
            p_TX_INT_DATAWIDTH           = 0,

            # CPLL Ports
            o_CPLLFBCLKLOST         = Open(),
            o_CPLLLOCK              = pll.lock if use_cpll else Open(),
            i_CPLLLOCKDETCLK        = 0,
            i_CPLLLOCKEN            = 1,
            i_CPLLPD                = 1 if use_qpll else 0,
            o_CPLLREFCLKLOST        = Open(),
            i_CPLLREFCLKSEL         = 0b001,
            i_CPLLRESET             = pll.reset if use_cpll else 0,
            i_GTRSVD                = 0b0000000000000000,
            i_PCSRSVDIN             = 0b0000000000000000,
            i_PCSRSVDIN2            = 0b00000,
            i_PMARSVDIN             = 0b00000,
            i_PMARSVDIN2            = 0b00000,
            i_TSTIN                 = 0b11111111111111111111,
            o_TSTOUT                = Open(10),

            # Channel
            i_CLKRSVD               = Cat(oob_clkdiv.o, 0, 0, 0),

            # Channel - Clocking Ports
            i_GTGREFCLK             = 0,
            i_GTNORTHREFCLK0        = 0,
            i_GTNORTHREFCLK1        = 0,
            i_GTREFCLK0             = pll._refclk if use_cpll else 0,
            i_GTREFCLK1             = 0,
            i_GTSOUTHREFCLK0        = 0,
            i_GTSOUTHREFCLK1        = 0,

            # Channel - DRP Ports
            i_DRPCLK                = ClockSignal("ss"),
            i_DRPADDR               = drp_arbiter.shared.addr,
            i_DRPDI                 = drp_arbiter.shared.di,
            o_DRPDO                 = drp_arbiter.shared.do,
            i_DRPWE                 = drp_arbiter.shared.we,
            i_DRPEN                 = drp_arbiter.shared.en,
            o_DRPRDY                = drp_arbiter.shared.rdy,

            # Clocking Ports
            o_GTREFCLKMONITOR       = Open(),
            i_QPLLCLK               = pll.clk    if use_qpll else 0,
            i_QPLLREFCLK            = pll.refclk if use_qpll else 0,
            i_RXSYSCLKSEL           = 0b11 if use_qpll else 0b00,
            i_TXSYSCLKSEL           = 0b11 if use_qpll else 0b00,

            # Digital Monitor Ports
            o_DMONITOROUT           = Open(8),

            # Loopback Ports
            i_LOOPBACK              = 0b000,

            # Power-Down Ports
            i_RXPD                  = 0b00,
            i_TXPD                  = 0b00,

            # RX Initialization and Reset Ports
            i_EYESCANRESET          = 0,
            i_GTRXRESET             = defer_rst.rx_o,
            i_RXOOBRESET            = 0,
            i_RXPCSRESET            = 0,
            i_RXPMARESET            = 0,
            o_RXRESETDONE           = rx_rst_done,
            i_RXUSERRDY             = 1,

            # Receive Ports - CDR Ports
            i_RXCDRFREQRESET        = 0,
            i_RXCDRHOLD             = 0,
            o_RXCDRLOCK             = Open(),
            i_RXCDROVRDEN           = 0,
            i_RXCDRRESET            = 0,
            i_RXCDRRESETRSV         = 0,

            # Receive Ports - Clock Correction Ports
            o_RXCLKCORCNT           = Open(2),

            # Receive Ports - FPGA RX Interface Datapath Configuration
            i_RX8B10BEN             = 1,

            # Receive Ports - FPGA RX Interface Ports
            i_RXUSRCLK              = self.pclk,
            i_RXUSRCLK2             = self.pclk,

            # Receive Ports - Pattern Checker Ports
            o_RXPRBSERR             = Open(),
            i_RXPRBSSEL             = 0b000,
            i_RXPRBSCNTRESET        = 0,

            # Receive Ports - PCI Express Ports
            o_PHYSTATUS             = Open(),
            i_RXRATE                = 0,
            o_RXSTATUS              = self.rx_status,
            o_RXVALID               = self.rx_valid,

            # Receive Ports - RX Data Path Interface
            o_RXDATA                = self.rx_data,

            # Receive Ports - RX 8B/10B Decoder Ports
            o_RXCHARISCOMMA         = Open(8),
            o_RXCHARISK             = self.rx_datak,
            o_RXDISPERR             = Open(8),
            o_RXNOTINTABLE          = Open(8),
            i_SETERRSTATUS          = 0,

            # Receive Ports - RX AFE Ports
            i_GTXRXN                = self._rx_pads.n,
            i_GTXRXP                = self._rx_pads.p,
            i_RXQPIEN               = 0,
            o_RXQPISENN             = Open(),
            o_RXQPISENP             = Open(),

            # Receive Ports - RX Buffer Bypass Ports
            i_RXBUFRESET            = 0,
            o_RXBUFSTATUS           = Open(3),
            i_RXDDIEN               = 0,
            i_RXDLYBYPASS           = 1,
            i_RXDLYEN               = 0,
            i_RXDLYOVRDEN           = 0,
            i_RXDLYSRESET           = 0,
            o_RXDLYSRESETDONE       = Open(),
            i_RXPHALIGN             = 0,
            o_RXPHALIGNDONE         = Open(),
            i_RXPHALIGNEN           = 0,
            i_RXPHDLYPD             = 0,
            i_RXPHDLYRESET          = 0,
            o_RXPHMONITOR           = Open(5),
            i_RXPHOVRDEN            = 0,
            o_RXPHSLIPMONITOR       = Open(5),

            # Receive Ports - RX Byte and Word Alignment Ports
            o_RXBYTEISALIGNED       = Open(),
            o_RXBYTEREALIGN         = Open(),
            o_RXCOMMADET            = Open(),
            i_RXCOMMADETEN          = 1,
            i_RXMCOMMAALIGNEN       = 1,
            i_RXPCOMMAALIGNEN       = 1,

            # Receive Ports - RX Channel Bonding Ports
            o_RXCHANBONDSEQ         = Open(),
            o_RXCHANISALIGNED       = Open(),
            o_RXCHANREALIGN         = Open(),
            i_RXCHBONDEN            = 0,
            i_RXCHBONDI             = 0b00000,
            i_RXCHBONDLEVEL         = 0b000,
            i_RXCHBONDMASTER        = 0,
            o_RXCHBONDO             = Open(5),
            i_RXCHBONDSLAVE         = 0,

            # Receive Ports - RX Equalizer Ports
            i_RXLPMEN               = 0,
            i_RXLPMHFHOLD           = ~self.rx_eq_training,
            i_RXLPMHFOVRDEN         = 0,
            i_RXLPMLFHOLD           = ~self.rx_eq_training,
            i_RXLPMLFKLOVRDEN       = 0,
            i_RXDFEAGCHOLD          = ~self.rx_eq_training,
            i_RXDFEAGCOVRDEN        = 0,
            i_RXDFECM1EN            = 0,
            i_RXDFELFHOLD           = ~self.rx_eq_training,
            i_RXDFELFOVRDEN         = 0,
            i_RXDFELPMRESET         = Open(),
            i_RXDFETAP2HOLD         = 0,
            i_RXDFETAP2OVRDEN       = 0,
            i_RXDFETAP3HOLD         = 0,
            i_RXDFETAP3OVRDEN       = 0,
            i_RXDFETAP4HOLD         = 0,
            i_RXDFETAP4OVRDEN       = 0,
            i_RXDFETAP5HOLD         = 0,
            i_RXDFETAP5OVRDEN       = 0,
            i_RXDFEUTHOLD           = 0,
            i_RXDFEUTOVRDEN         = 0,
            i_RXDFEVPHOLD           = 0,
            i_RXDFEVPOVRDEN         = 0,
            i_RXDFEVSEN             = 0,
            i_RXDFEXYDEN            = 1,
            i_RXDFEXYDHOLD          = 0,
            i_RXDFEXYDOVRDEN        = 0,
            o_RXMONITOROUT          = Open(7),
            i_RXMONITORSEL          = 0b00,
            i_RXOSHOLD              = 0,
            i_RXOSOVRDEN            = 0,

            # Receive Ports - RX Fabric Clock Output Control Ports
            o_RXRATEDONE            = Open(),

            # Receive Ports - RX Fabric Output Control Ports
            o_RXOUTCLK              = Open(),
            o_RXOUTCLKFABRIC        = Open(),
            o_RXOUTCLKPCS           = Open(),
            i_RXOUTCLKSEL           = 0b010,

            # Receive Ports - RX Gearbox Ports
            o_RXDATAVALID           = Open(),
            o_RXHEADER              = Open(3),
            o_RXHEADERVALID         = Open(1),
            o_RXSTARTOFSEQ          = Open(),
            i_RXGEARBOXSLIP         = 0,
            i_RXSLIDE               = 0,

            # Receive Ports - RX Margin Analysis Ports
            o_EYESCANDATAERROR      = Open(),
            i_EYESCANMODE           = 0,
            i_EYESCANTRIGGER        = 0,

            # Receive Ports - RX OOB Signaling Ports
            o_RXCOMSASDET           = Open(),
            o_RXCOMWAKEDET          = Open(),
            o_RXCOMINITDET          = Open(),
            o_RXELECIDLE            = self.rx_elec_idle,
            i_RXELECIDLEMODE        = 0b00,

            # Receive Ports - RX Polarity Control Ports
            i_RXPOLARITY            = self.rx_polarity,

            # TX Initialization and Reset Ports
            i_CFGRESET              = 0,
            i_GTTXRESET             = defer_rst.tx_o,
            i_TXPCSRESET            = 0,
            i_TXPMARESET            = 0,
            o_TXRESETDONE           = tx_rst_done,
            i_TXUSERRDY             = 1,
            o_PCSRSVDOUT            = Open(16),

            # Transceiver Reset Mode Operation
            i_GTRESETSEL            = 0,
            i_RESETOVRD             = 0,

            # Transmit Ports - Configurable Driver Ports
            o_GTXTXN                = self._tx_pads.n,
            o_GTXTXP                = self._tx_pads.p,
            i_TXBUFDIFFCTRL         = 0b100,
            i_TXDEEMPH              = 0,
            i_TXDIFFCTRL            = 0b1000,
            i_TXDIFFPD              = 0,
            i_TXINHIBIT             = self.tx_gpio_en,
            i_TXMAINCURSOR          = 0b0000000,
            i_TXPISOPD              = 0,
            i_TXPOSTCURSOR          = 0b00000,
            i_TXPOSTCURSORINV       = 0,
            i_TXPRECURSOR           = 0b00000,
            i_TXPRECURSORINV        = 0,
            o_TXQPISENN             = Open(),
            o_TXQPISENP             = Open(),
            i_TXQPIBIASEN           = 0,
            i_TXQPISTRONGPDOWN      = 0,
            i_TXQPIWEAKPUP          = 0,

            # Transmit Ports - FPGA TX Interface Datapath Configuration
            i_TX8B10BEN             = 1,

            # Transmit Ports - FPGA TX Interface Ports
            i_TXUSRCLK              = self.pclk,
            i_TXUSRCLK2             = self.pclk,

            # Transmit Ports - PCI Express Ports
            i_TXELECIDLE            = ~self.tx_gpio_en & self.tx_elec_idle,
            i_TXMARGIN              = 0,
            i_TXRATE                = 0b000,
            i_TXSWING               = 0,

            # Transmit Ports - Pattern Generator Ports
            i_TXPRBSSEL             = 0b000,
            i_TXPRBSFORCEERR        = 0,

            # Transmit Ports - TX 8B/10B Encoder Ports
            i_TX8B10BBYPASS         = 0b00000000,
            i_TXCHARDISPMODE        = 0b00000000,
            i_TXCHARDISPVAL         = 0b00000000,
            i_TXCHARISK             = self.tx_datak,

            # Transmit Ports - TX Data Path Interface
            i_TXDATA                = self.tx_data,

            # Transmit Ports - TX Buffer Bypass Ports
            i_TXDLYBYPASS           = 1,
            i_TXDLYEN               = 0,
            i_TXDLYHOLD             = 0,
            i_TXDLYOVRDEN           = 0,
            i_TXDLYSRESET           = 0,
            o_TXDLYSRESETDONE       = Open(),
            i_TXDLYUPDOWN           = 0,
            i_TXPHALIGN             = 0,
            o_TXPHALIGNDONE         = Open(),
            i_TXPHALIGNEN           = 0,
            i_TXPHDLYPD             = 0,
            i_TXPHDLYRESET          = 0,
            i_TXPHDLYTSTCLK         = 0,
            i_TXPHINIT              = 0,
            o_TXPHINITDONE          = Open(),
            i_TXPHOVRDEN            = 0,

            # Transmit Ports - TX Buffer Ports
            o_TXBUFSTATUS           = Open(2),

            # Transmit Ports - TX Fabric Clock Output Control Ports
            o_TXOUTCLK              = txoutclk,
            o_TXOUTCLKFABRIC        = Open(),
            o_TXOUTCLKPCS           = Open(),
            i_TXOUTCLKSEL           = 0b010,
            o_TXRATEDONE            = Open(),

            # Transmit Ports - TX Gearbox Ports
            o_TXGEARBOXREADY        = Open(),
            i_TXHEADER              = 0b000,
            i_TXSEQUENCE            = 0b0000000,
            i_TXSTARTSEQ            = 0,

            # Transmit Ports - TX OOB Signalling Ports
            o_TXCOMFINISH           = Open(),
            i_TXCOMINIT             = 0,
            i_TXCOMSAS              = 0,
            i_TXCOMWAKE             = 0,
            i_TXPDELECIDLEMODE      = 0,

            # Transmit Ports - TX Polarity Control Ports
            i_TXPOLARITY            = self.tx_polarity ^ (self.tx_gpio_en & self.tx_gpio),

            # Transmit Ports - TX Receiver Detection Ports
            i_TXDETECTRX            = 0,
        )


        return m



class XC7GTXSerDesPIPE(PIPEInterface, Elaboratable):
    """ Wrapper around the core GTX SerDes that adapts it to the PIPE interface.

    The implementation-dependent behavior of the standard PIPE signals is described below:

    width :
        Interface width. Always 2 symbols.
    clk :
        Reference clock for the PHY receiver and transmitter. Could be routed through fabric,
        or connected to the output of an ``IBUFDS_GTE2`` block.
    pclk :
        Clock for the PHY interface. Frequency is always 250 MHz.
    phy_mode :
        PHY operating mode. Only SuperSpeed USB mode is supported.
    elas_buf_mode :
        Elastic buffer mode. Only nominal half-full mode is supported.
    rate :
        Link signaling rate. Only 5 GT/s is supported.
    power_down :
        Power management mode. Only P0 is supported.
    tx_deemph :
        Transmitter de-emphasis level. Only TBD is supported.
    tx_margin :
        Transmitter voltage levels. Only TBD is supported.
    tx_swing :
        Transmitter voltage swing level. Only full swing is supported.
    tx_detrx_lpbk :
    tx_elec_idle :
        Transmit control signals. Loopback and receiver detection are not implemented.
    tx_compliance :
    tx_ones_zeroes :
        These inputs are not implemented.
    power_present :
        This output is not implemented. External logic may drive it if necessary.
    """

    def __init__(self, *, tx_pads, rx_pads, refclk_frequency, ss_clock_frequency):
        super().__init__(width=2)

        self._tx_pads            = tx_pads
        self._rx_pads            = rx_pads
        self._refclk_frequency   = refclk_frequency
        self._ss_clock_frequency = ss_clock_frequency


    def elaborate(self, platform):
        m = Module()

        #
        # PLL and SerDes instantiation.
        #
        m.submodules.pll = pll = GTXQuadPLL(
            refclk              = self.clk,
            refclk_freq         = self._refclk_frequency,
            linerate            = 5e9
        )
        m.submodules.serdes = serdes = GTXChannel(
            pll                 = pll,
            tx_pads             = self._tx_pads,
            rx_pads             = self._rx_pads,
            ss_clock_frequency  = self._ss_clock_frequency
        )

        # Our soft PHY includes some logic that needs to run synchronously to the PIPE clock; create
        # a local clock domain to drive it.
        m.domains.pipe = ClockDomain(local=True, async_reset=True)
        m.d.comb += [
            ClockSignal("pipe")     .eq(serdes.pclk),
        ]


        #
        # LFPS generation.
        #
        m.submodules.lfps_generator = lfps_generator = LFPSSquareWaveGenerator(25e6, 250e6)
        m.d.comb += [
            serdes.tx_gpio_en       .eq(lfps_generator.tx_gpio_en),
            serdes.tx_gpio          .eq(lfps_generator.tx_gpio),
        ]


        #
        # PIPE interface signaling.
        #
        m.d.comb += [
            pll.reset               .eq(self.reset),
            serdes.reset            .eq(self.reset),
            self.pclk               .eq(serdes.pclk),

            serdes.tx_elec_idle     .eq(self.tx_elec_idle),
            serdes.rx_polarity      .eq(self.rx_polarity),
            serdes.rx_eq_training   .eq(self.rx_eq_training),
            serdes.rx_termination   .eq(self.rx_termination),
            lfps_generator.generate .eq(self.tx_detrx_lpbk & self.tx_elec_idle),

            self.phy_status         .eq(~serdes.tx_ready),
            self.rx_valid           .eq(serdes.rx_valid),
            self.rx_status          .eq(serdes.rx_status),
            self.rx_elec_idle       .eq(serdes.rx_elec_idle),

            serdes.tx_data          .eq(self.tx_data),
            serdes.tx_datak         .eq(self.tx_datak),
            self.rx_data            .eq(serdes.rx_data),
            self.rx_datak           .eq(serdes.rx_datak),
        ]

        return m
